#!/usr/bin/env python
# Portions Copyright 2012 Google Inc. All Rights Reserved.
# This file has been modified by NICTA
import phe.encoding
from phe.paillier import PaillierPrivateKey, PaillierPublicKey

# This file is part of pyphe.
#
# Pyphe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyphe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyphe.  If not, see <http://www.gnu.org/licenses/>.

"""Unittest for paillier module."""

import logging
import unittest
import sys
import math

from phe import paillier


class PaillierGeneric(unittest.TestCase):

    def testDefaultCreateKeypair(self):
        public_key, private_key = paillier.generate_paillier_keypair()

        self.assertTrue(hasattr(public_key, 'g'))
        self.assertTrue(hasattr(public_key, 'n'))

        self.assertTrue(hasattr(private_key, 'p'))
        self.assertTrue(hasattr(private_key, 'q'))
        self.assertTrue(hasattr(private_key, 'public_key'))

        self.assertTrue(str(public_key).startswith('<PaillierPublicKey '))
        self.assertTrue(str(private_key).startswith('<PaillierPrivateKey '))

    def testCreateKeypairLengths(self):

        for key_length in [8, 16, 32, 64, 128, 256, 512, 1024, 2048, 3072, 4096]:
            public_key, private_key = paillier.generate_paillier_keypair(n_length=key_length)

            self.assertTrue(hasattr(public_key, 'g'))
            self.assertTrue(hasattr(public_key, 'n'))

            self.assertTrue(hasattr(private_key, 'p'))
            self.assertTrue(hasattr(private_key, 'q'))

            # Check that no exceptions are raised representing these keys
            repr(public_key)
            repr(private_key)

    def testKeyUniqueness(self):
        repeats = 100
        public_keys = set()
        private_keys = set()
        for _ in range(repeats):
            public_key, private_key = paillier.generate_paillier_keypair(n_length=256)
            self.assertNotIn(public_key, public_keys, "Managed to generate the same public key")
            self.assertNotIn(private_key, private_keys, "Managed to generate the same private key")
            public_keys.add(public_key)
            private_keys.add(private_key)

    def testStaticPrivateKeyConstructor(self):
        public_key, private_key = paillier.generate_paillier_keypair()
        p = private_key.p
        q = private_key.q
        private_key_from_static = PaillierPrivateKey.from_totient(public_key, (p-1) * (q-1))
        c = public_key.encrypt(4242)
        self.assertEqual(private_key, private_key_from_static, "The private keys should be the same.")
        self.assertEqual(private_key_from_static.decrypt(c), 4242, "Result of the decryption should be 4242")

    def testPrivateKeyEquality(self):
        pk = PaillierPublicKey(2537)
        p1 = PaillierPrivateKey(pk, 43, 59)
        p2 = PaillierPrivateKey(pk, 59, 43)
        self.assertEqual(p1, p2, "These private keys should be equal")

class PaillierTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Could move this into setUpModule() if we get too many classes
        cls.public_key, cls.private_key = paillier.generate_paillier_keypair()
        cls.other_public_key, cls.other_private_key = paillier.generate_paillier_keypair()

    def assertInRange(self, value, a, b):
        """Assert that a <= value <= b or a >= value >= b.

        Useful when python floats suffer rounding errors and Paillier
        floats are at least as good as the python floats
        """
        if not ((a <= value <= b) or (a >= value >= b)):
            self.fail("%s is not between %s and %s" % (value, a, b))


class PaillierTestRawEncryption(PaillierTest):

    def testEncryptDecrypt(self):
        data = 123456789123456789123456789123456789
        ciphertext = self.public_key.raw_encrypt(data)
        decryption = self.private_key.raw_decrypt(ciphertext)
        self.assertEqual(data, decryption)

    def testModuloN(self):
        # Check decryption works for n -1
        plaintext1 = self.public_key.n - 1
        ciphertext1 = self.public_key.raw_encrypt(plaintext1)
        self.assertEqual(plaintext1, self.private_key.raw_decrypt(ciphertext1))
        # Check decryption wraps for n to 0
        plaintext2 = self.public_key.n
        ciphertext2 = self.public_key.raw_encrypt(plaintext2)
        self.assertEqual(0, self.private_key.raw_decrypt(ciphertext2))
        # Check decryption wraps for n + 1 to 1
        plaintext3 = self.public_key.n + 1
        ciphertext3 = self.public_key.raw_encrypt(plaintext3)
        self.assertEqual(1, self.private_key.raw_decrypt(ciphertext3))

    def testRawEncryptDecryptRegression0(self):

        public_key = paillier.PaillierPublicKey(126869)
        private_key = paillier.PaillierPrivateKey(public_key, 293, 433)

        ciphertext = public_key.raw_encrypt(10100, 74384)
        self.assertEqual(935906717, ciphertext)
        decryption = private_key.raw_decrypt(935906717)
        self.assertEqual(10100, decryption)

    def testEncryptRegression(self):
        public_key = paillier.PaillierPublicKey(126869)

        enc_num = public_key.encrypt(10100, r_value=74384)
        self.assertEqual(935906717, enc_num.ciphertext(False))

    def testEncryptIsRandom(self):
        # Check for semantic security
        public_key = paillier.PaillierPublicKey(126869)

        enc_num = public_key.encrypt(1, r_value=1)
        self.assertEqual(126870, enc_num.ciphertext(False))
        # r_value should be random
        enc_num2 = public_key.encrypt(1)
        enc_num3 = public_key.encrypt(1)
        self.assertNotEqual(126870, enc_num2.ciphertext(False))
        self.assertNotEqual(enc_num2.ciphertext(False),
                            enc_num3.ciphertext(False))

    def testEncryptInvalidType(self):
        data = "123"
        self.assertRaises(TypeError, self.public_key.raw_encrypt, data)

    def testDecryptInvalidType(self):
        data = 123456789123456789123456789123456789
        ciphertext = self.public_key.raw_encrypt(data)
        self.assertRaises(TypeError, self.private_key.raw_decrypt, str(ciphertext))


class PaillierTestEncodedNumber(PaillierTest):


    def setUp(self):
        super().setUp()
        self.EncodedNumberCls = phe.encoding.EncodedNumber


    def testEncodeInt0(self):
        # A small positive number
        enc = self.EncodedNumberCls.encode(self.public_key, 15)
        self.assertEqual(0, enc.exponent)
        self.assertEqual(15, enc.encoding)

    def testEncodeInt1(self):
        # A small negative number
        enc = self.EncodedNumberCls.encode(self.public_key, -15)
        self.assertEqual(0, enc.exponent)
        self.assertNotEqual(-15, enc.encoding)
        self.assertEqual(-15 % self.public_key.n, enc.encoding)

    def testDecodeInt0(self):
        enc = self.EncodedNumberCls(self.public_key, 15, 0)
        self.assertEqual(15, enc.decode())

    def testDecodeInt1(self):
        enc = self.EncodedNumberCls(self.public_key, -15 % self.public_key.n, 0)
        self.assertEqual(-15, enc.decode())

    def testEncodeIntDecodeInt2(self):
        # large positive number
        encoded = self.EncodedNumberCls.encode(self.public_key, 2 ** 140)
        self.assertEqual(0, encoded.exponent)
        decoded = encoded.decode()
        self.assertEqual(2 ** 140, decoded)

    def testEncodeIntDecodeInt3(self):
        # large negative number
        encoded = self.EncodedNumberCls.encode(self.public_key, -2 ** 140)
        self.assertEqual(0, encoded.exponent)
        decoded = encoded.decode()
        self.assertEqual(-2 ** 140, decoded)

    def testEncodeIntDecodeInt4(self):
        # largest positive number
        encoded = self.EncodedNumberCls.encode(self.public_key,
                                                self.public_key.max_int)
        self.assertEqual(0, encoded.exponent)
        decoded = encoded.decode()
        self.assertEqual(self.public_key.max_int, decoded)

    def testEncodeIntDecodeInt5(self):
        # largest negative number
        encoded = self.EncodedNumberCls.encode(self.public_key,
                                                -self.public_key.max_int)
        decoded = encoded.decode()
        self.assertEqual(-self.public_key.max_int, decoded)

    def testEncodeIntTooLargePositive(self):
        # check value error is raised on too large a positive input
        self.assertRaises(ValueError, self.EncodedNumberCls.encode,
                          self.public_key, self.public_key.max_int + 1)
        self.assertRaises(ValueError, self.EncodedNumberCls.encode,
                          self.public_key, 2 ** (paillier.DEFAULT_KEYSIZE-1))

    def testEncodeIntTooLargeNegative(self):
        # check value error is raised on too large a positive input
        self.assertRaises(ValueError, self.EncodedNumberCls.encode,
                          self.public_key, -self.public_key.max_int - 1)
        self.assertRaises(ValueError, self.EncodedNumberCls.encode,
                          self.public_key, -2 ** (paillier.DEFAULT_KEYSIZE-1))

    def testDecodeCorruptEncodedNumber(self):
        encoded = self.EncodedNumberCls.encode(self.public_key, 10)
        encoded.encoding += self.public_key.n
        self.assertRaises(ValueError, encoded.decode)

    def testDecodeWithOverflowEncodedNumber(self):
        encoded = self.EncodedNumberCls.encode(self.public_key, 10)
        encoded.encoding += self.public_key.max_int
        self.assertRaises(OverflowError, encoded.decode)

    def testEncodeFloat0(self):
        enc = self.EncodedNumberCls.encode(self.public_key, 15.1)
        negexp = self.EncodedNumberCls.BASE ** enc.exponent
        dec = self.EncodedNumberCls.BASE ** enc.exponent * enc.encoding
        self.assertAlmostEqual(15.1, dec)

    def testEncodeFloatDecodeFloat0(self):
        enc = self.EncodedNumberCls.encode(self.public_key, 15.1)
        self.assertAlmostEqual(15.1, enc.decode())

    def testEncodeFloatDecodeFloat1(self):
        enc = self.EncodedNumberCls.encode(self.public_key, -15.1)
        self.assertAlmostEqual(-15.1, enc.decode())

    def testEncryptFloatDecryptFloat2(self):
        # large positive number
        encoded = self.EncodedNumberCls.encode(self.public_key, 2.1 ** 20)
        self.assertEqual(2.1 ** 20, encoded.decode())

        encrypted = self.public_key.encrypt(encoded)

        decrypted_but_encoded = self.private_key.decrypt_encoded(encrypted, self.EncodedNumberCls)

        self.assertAlmostEqual(2.1 ** 20, decrypted_but_encoded.decode())

    def testEncryptFloatDecryptFloat3(self):
        # large negative number
        encoded = self.EncodedNumberCls.encode(self.public_key, -2.1 ** 63)
        self.assertAlmostEqual(-2.1 ** 63, encoded.decode())
        encrypted = self.public_key.encrypt(encoded)
        decrypted_but_encoded = self.private_key.decrypt_encoded(encrypted, self.EncodedNumberCls)
        self.assertAlmostEqual(-2.1 ** 63, decrypted_but_encoded.decode())

    def testEncodedDecreaseExponentTo0(self):
        # Check that decrease_exponent_to does what it says
        enc1 = self.EncodedNumberCls.encode(self.public_key, 3.14)
        new_exponent = enc1.exponent - 10
        assert new_exponent < enc1.exponent # So the next part is meaningful
        enc2 = enc1.decrease_exponent_to(new_exponent)

        self.assertLess(new_exponent, enc1.exponent)
        self.assertEqual(new_exponent, enc2.exponent)
        self.assertAlmostEqual(3.14, enc2.decode())

    def testEncodedDecreaseExponentTo1(self):
        # Check that decrease_exponent_to does what it says
        enc1 = self.EncodedNumberCls.encode(self.public_key, -3.14)
        new_exponent = enc1.exponent - 10

        assert new_exponent < enc1.exponent # So the next part is meaningful
        enc2 = enc1.decrease_exponent_to(new_exponent)
        self.assertLess(new_exponent, enc1.exponent)
        self.assertEqual(new_exponent, enc2.exponent)
        self.assertAlmostEqual(-3.14, enc2.decode())

    def testEncodedDecreaseInvalidExponent(self):
        # Check that decrease_exponent_to does what it says
        enc1 = self.EncodedNumberCls.encode(self.public_key, 3.14)
        assert enc1.exponent < -8
        self.assertRaises(ValueError, enc1.decrease_exponent_to, -8)


class PaillierTestEncodedNumberDefaultBase(PaillierTestEncodedNumber):
    """Encoded Number tests with a default encoding base.
    """

    def testManualPrecision0(self):
        # Check that the encoded +ve number is precise enough...
        val, prec = 3.171234e-7, 1e-8
        encoding = self.EncodedNumberCls.encode(self.public_key, val, precision=prec)
        decoded = encoding.decode()
        self.assertInRange(decoded, val - prec, val + prec)

        # Well, that didn't actually prove much - what if val happens
        # to be conveniently representable in BASE?
        # `decoded` *is* conveniently representable in BASE, so let's
        # play with that a little
        encoding2 = self.EncodedNumberCls.encode(self.public_key,
                                                  decoded + 0.500001 * prec,
                                                  precision=prec)
        decoded2 = encoding2.decode()
        self.assertNotEqual(decoded, decoded2)
        self.assertInRange(decoded2, val - prec/2, val + prec*1.5001)

        # Check it's not too precise:
        val3 = decoded + prec / self.EncodedNumberCls.BASE
        encoding3 = self.EncodedNumberCls.encode(self.public_key, val3, precision=prec)
        decoded3 = encoding3.decode()
        self.assertEqual(decoded, decoded3)

    def testManualPrecision1(self):
        # Check that the encoded -ve number is precise enough...
        val, prec = -3.171234e-7, 1e-8
        encoding = self.EncodedNumberCls.encode(self.public_key, val, precision=prec)
        decoded = encoding.decode()
        self.assertInRange(decoded, val - prec, val + prec)

        # Well, that didn't actually prove much - what if val happens
        # to be conveniently representable in BASE?
        # `decoded` *is* conveniently representable in BASE, so let's
        # play with that a little
        encoding2 = self.EncodedNumberCls.encode(self.public_key,
                                                  decoded + 0.500001 * prec,
                                                  precision=prec)
        decoded2 = encoding2.decode()
        self.assertNotEqual(decoded, decoded2)
        self.assertInRange(decoded2, val, val + prec)

        # Check it's not too precise:
        val3 = decoded + prec / self.EncodedNumberCls.BASE
        encoding3 = self.EncodedNumberCls.encode(self.public_key, val3, precision=prec)
        decoded3 = encoding3.decode()
        self.assertEqual(decoded, decoded3)

    def testAutomaticPrecisionAgreesWithEpsilon(self):
        # Check that automatic precision is equivalent to precision=eps
        eps = sys.float_info.epsilon

        # There's a math.floor in _encode, we want to test that
        # bin_lsb_exponent is correct and not off by some fraction that
        # sometimes gets rounded down. The " * 2" in the next line is excessive.
        floor_happy = math.ceil(self.EncodedNumberCls.LOG2_BASE) * 2

        for i in range(-floor_happy, floor_happy + 1):
            enc1 = self.EncodedNumberCls.encode(self.public_key, 2.**i)
            enc2 = self.EncodedNumberCls.encode(self.public_key, 2.**i,
                                                          precision=eps * 2**i)
            self.assertEqual(enc1.exponent, enc2.exponent, i)

            # Check the max val for a given eps
            rel_eps = eps * 2 ** (i - 1)
            val = 2. ** i - rel_eps
            assert val != 2. ** i
            enc3 = self.EncodedNumberCls.encode(self.public_key, val)
            enc4 = self.EncodedNumberCls.encode(self.public_key, val,
                                                        precision=rel_eps)
            self.assertEqual(enc3.exponent, enc4.exponent, i)


class PaillierTestEncodedNumberAlternativeBaseLarge(PaillierTestEncodedNumber):
    """Encoded Number tests with a different encoding base.
    """

    def setUp(self):
        super().setUp()

        class AltEncodedNumber(phe.encoding.EncodedNumber):
            BASE = 64
            LOG2_BASE = math.log(BASE, 2)

        self.EncodedNumberCls = AltEncodedNumber


class PaillierTestEncodedNumberAlternativeBaseSmall(PaillierTestEncodedNumber):
    """Encoded Number tests with a different encoding base.
    """

    def setUp(self):
        super().setUp()

        class AltEncodedNumber(phe.encoding.EncodedNumber):
            BASE = 2
            LOG2_BASE = math.log(BASE, 2)

        self.EncodedNumberCls = AltEncodedNumber


class PaillierTestEncodedNumberAlternativeBaseOdd(PaillierTestEncodedNumber):
    """Encoded Number tests with an odd encoding base.
    """

    def setUp(self):
        super().setUp()

        class AltEncodedNumber(phe.encoding.EncodedNumber):
            BASE = 13
            LOG2_BASE = math.log(BASE, 2)

        self.EncodedNumberCls = AltEncodedNumber


class PaillierTestEncryptedNumber(PaillierTest):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.p_n = [123456789, 314159265359,
                    271828182846, -987654321,
                    -161803398874, -1414213562373095]
        cls.c_n = [cls.public_key.encrypt(n) for n in cls.p_n]

    def testCreateEncryptedNumber(self):
        paillier.EncryptedNumber(self.public_key, 5)

    def testEncryptIntDecryptInt0(self):
        # A small positive number
        ciphertext = self.public_key.encrypt(15)
        decryption = self.private_key.decrypt(ciphertext)
        self.assertEqual(15, decryption)

    def testEncryptIntDecryptInt1(self):
        # A small negative number
        ciphertext = self.public_key.encrypt(-15)
        decryption = self.private_key.decrypt(ciphertext)
        self.assertEqual(-15, decryption)

    def testEncryptIntDecryptInt4(self):
        # largest positive number
        ciphertext = self.public_key.encrypt(self.public_key.max_int)
        decryption = self.private_key.decrypt(ciphertext)
        self.assertEqual(self.public_key.max_int, decryption)

    def testEncryptIntDecryptInt5(self):
        # largest negative number
        ciphertext = self.public_key.encrypt(-self.public_key.max_int)
        decryption = self.private_key.decrypt(ciphertext)
        self.assertEqual(-self.public_key.max_int, decryption)

    def testCantEncryptDecryptIntWithDifferentKey(self):
        data = 1564
        ciphertext = self.public_key.encrypt(data)
        self.assertRaises(ValueError, self.other_private_key.decrypt, ciphertext)

    def testCantEncryptDecryptIntWithDifferentSizeKey(self):
        public_key, private_key = paillier.generate_paillier_keypair(n_length=128)
        data = 1564
        ciphertext = self.public_key.encrypt(data)
        self.assertRaises(ValueError, private_key.decrypt, ciphertext)

    def testCantAddWithDifferentKey(self):
        ciphertext1 = self.public_key.encrypt(-15)
        ciphertext2 = self.other_public_key.encrypt(1)
        self.assertRaises(ValueError, ciphertext1.__add__, ciphertext2)

    def testCantAddEncodedWithDifferentKey(self):
        ciphertext1 = self.public_key.encrypt(-15)
        ciphertext2 = phe.encoding.EncodedNumber(self.other_public_key, 1, ciphertext1.exponent)
        self.assertRaises(ValueError, ciphertext1.__add__, ciphertext2)

    def testAddWithEncryptDecryptInt0(self):
        # Add 1 to a small negative number
        ciphertext1 = self.public_key.encrypt(-15)
        ciphertext2 = self.public_key.encrypt(1)
        ciphertext3 = ciphertext1 + ciphertext2
        decryption = self.private_key.decrypt(ciphertext3)
        self.assertEqual(-14, decryption)

    def testAddWithEncryptDecryptInt1(self):
        # Add 1 to a small positive number
        ciphertext1 = self.public_key.encrypt(15)
        ciphertext2 = self.public_key.encrypt(1)
        ciphertext3 = ciphertext1 + ciphertext2
        decryption = self.private_key.decrypt(ciphertext3)
        self.assertEqual(16, decryption)

    def testAddWithEncryptDecryptInt2(self):
        # Add -1 to a small negative number
        ciphertext1 = self.public_key.encrypt(-15)
        ciphertext2 = self.public_key.encrypt(-1)
        ciphertext3 = ciphertext1 + ciphertext2
        decryption = self.private_key.decrypt(ciphertext3)
        self.assertEqual(-16, decryption)

    def testSubWithEncryptDecryptInt0(self):
        # Subtract two encrypted integers
        ciphertext1 = self.public_key.encrypt(15)
        ciphertext2 = self.public_key.encrypt(1)
        ciphertext3 = ciphertext1 - ciphertext2
        decryption = self.private_key.decrypt(ciphertext3)
        self.assertEqual(14, decryption)

    def testSubScalarWithEncryptDecryptInt0(self):
        # Subtract two encrypted integers
        ciphertext1 = self.public_key.encrypt(15)
        ciphertext2 = ciphertext1 - 2
        decryption = self.private_key.decrypt(ciphertext2)
        self.assertEqual(13, decryption)

    def testSubScalarWithEncryptDecryptInt0Right(self):
        # Subtract two encrypted integers
        ciphertext1 = self.public_key.encrypt(15)
        ciphertext2 = 20 - ciphertext1
        decryption = self.private_key.decrypt(ciphertext2)
        self.assertEqual(5, decryption)

    def testMultipleAddWithEncryptDecryptInt0(self):
        # Add many positive numbers
        ciphertext = self.c_n[0] + self.c_n[1] + self.c_n[2]
        decryption = self.private_key.decrypt(ciphertext)
        self.assertEqual(sum(self.p_n[0:3]), decryption)

    def testMultipleAddWithEncryptDecryptInt1(self):
        # Add many negative numbers
        ciphertext = self.c_n[3] + self.c_n[4] + self.c_n[5]
        decryption = self.private_key.decrypt(ciphertext)
        self.assertEqual(sum(self.p_n[3:6]), decryption)

    def testMultipleAddWithEncryptDecryptInt2(self):
        # Add many positive and negative numbers with aggregate being positive
        ciphertext = sum(self.c_n[:5])
        decryption = self.private_key.decrypt(ciphertext)
        sum_3pos_2neg = sum(self.p_n[:5])
        self.assertEqual(sum_3pos_2neg, decryption)

    def testMultipleAddWithEncryptDecryptInt3(self):
        # Add many positive and negative numbers with aggregate being negative
        ciphertext = sum(self.c_n)
        decryption = self.private_key.decrypt(ciphertext)
        sum_3pos_3neg = sum(self.p_n)
        self.assertEqual(sum_3pos_3neg, decryption)

    def testMultipleAddWithEncryptDecryptIntLimits(self):
        sum_3pos_2neg = sum(self.p_n[:5])
        sum_3pos_3neg = sum(self.p_n)
        ciphertext_3pos_2neg = sum(self.c_n[:5])
        ciphertext_3pos_3neg = sum(self.c_n)

        # Add many positive and negative numbers to reach max_int.
        ciphertext = self.public_key.encrypt(self.public_key.max_int - sum_3pos_2neg)
        ciphertext = ciphertext_3pos_2neg + ciphertext
        decryption = self.private_key.decrypt(ciphertext)
        self.assertEqual(self.public_key.max_int, decryption)

        # Add many positive and negative numbers to reach -max_int.
        ciphertext = self.public_key.encrypt(-self.public_key.max_int - sum_3pos_3neg)
        ciphertext = ciphertext_3pos_3neg + ciphertext
        decryption = self.private_key.decrypt(ciphertext)
        self.assertEqual(-self.public_key.max_int, decryption)


    def testAffineWithEncryptDecryptInt(self):
        logging.debug('Running testAffineWithEncryptDecryptInt method.')
        plaintext1 = 123456789123456789123456789123456789
        for plaintext in (plaintext1, -plaintext1):
            ciphertext1 = self.public_key.encrypt(plaintext)
            # tests a = 2
            a = 2
            b = 111111110111111110111111110111111110
            ciphertext3 = ciphertext1 * a + b
            decryption3 = self.private_key.decrypt(ciphertext3)
            self.assertEqual(a * plaintext + b, decryption3)
            # tests a = 0
            ciphertext4 = ciphertext1 * 0 + b
            decryption4 = self.private_key.decrypt(ciphertext4)
            self.assertEqual(b, decryption4)
            # tests a = 1
            ciphertext5 = ciphertext1 * 1 + b
            decryption5 = self.private_key.decrypt(ciphertext5)
            self.assertEqual(plaintext + b, decryption5)
            # tests b = 0
            ciphertext6 = ciphertext1 * 2 + 0
            decryption6 = self.private_key.decrypt(ciphertext6)
            self.assertEqual(2 * plaintext, decryption6)
            # tests a=0, b = 0
            ciphertext7 = ciphertext1 * 0 + 0
            decryption7 = self.private_key.decrypt(ciphertext7)
            self.assertEqual(0, decryption7)

    def testEncryptIntPositiveOverflowAdd(self):
        # positive overflow as a result of adding
        ciphertext1 = self.public_key.encrypt(self.public_key.max_int)
        ciphertext2 = self.public_key.encrypt(1)
        ciphertext3 = ciphertext1 + ciphertext2
        self.assertRaises(OverflowError, self.private_key.decrypt, ciphertext3)

    def testEncryptIntNegativeOverflowAdd(self):
        # negative overflow as a result of adding
        ciphertext1 = self.public_key.encrypt(-self.public_key.max_int)
        ciphertext2 = self.public_key.encrypt(-1)
        ciphertext3 = ciphertext1 + ciphertext2
        self.assertRaises(OverflowError, self.private_key.decrypt,ciphertext3)

    def testUndetectableAddOverflow(self):
        # Prove the ring does wrap
        ciphertext1 = self.public_key.encrypt(0)
        ciphertext2 = self.public_key.encrypt(self.public_key.max_int)
        ciphertext3 = self.public_key.encrypt(self.public_key.max_int)
        ciphertext4 = self.public_key.encrypt(self.public_key.max_int)
        cipher_sum = ciphertext1 + ciphertext2 + ciphertext3 + ciphertext4

        plain_sum = self.private_key.decrypt(cipher_sum)
        # plain_sum = 3 * max_int = 3 * ((n//3) - 1)
        # due to residues of the // function,
        # -5 < plain_sum < -3 (modulo n)
        self.assertLessEqual(-5, plain_sum)
        self.assertGreaterEqual(-3, plain_sum)

    def testCantAddWithDifferentKeys(self):
        ciphertext1 = self.public_key.encrypt(0, r_value=1)
        # Let's not and say we did
        ciphertext2 = self.public_key.encrypt(20, r_value=1)
        public_key_2 = paillier.PaillierPublicKey(126869)
        ciphertext2.public_key = public_key_2  # Suuuper dodgy

        self.assertRaises(ValueError, ciphertext1.__add__, ciphertext2)

    def testAddWithEncryptedIntAndEncodedNumber(self):
        # Add 1 to a small positive number
        ciphertext1 = self.public_key.encrypt(15)
        encoded2 = phe.encoding.EncodedNumber.encode(self.public_key, 1)
        ciphertext3 = ciphertext1 + encoded2
        decryption = self.private_key.decrypt(ciphertext3)
        self.assertEqual(16, decryption)

    def testAddWithEncryptedIntAndEncodedNumberDiffExp0(self):
        # Add 1 to a small positive number
        ciphertext1 = self.public_key.encrypt(15)
        encoded2 = phe.encoding.EncodedNumber.encode(self.public_key, 1, max_exponent=-50)
        assert encoded2.exponent > -200
        assert ciphertext1.exponent > -200

        encoded2 = encoded2.decrease_exponent_to(-200)
        ciphertext3 = ciphertext1 + encoded2
        self.assertEqual(16, self.private_key.decrypt(ciphertext3))

    def testAddWithEncryptedIntAndEncodedNumberDiffExp1(self):
        # Try with the EncryptedNumber having the smaller exponent
        ciphertext1 = self.public_key.encrypt(15)
        ciphertext2 = ciphertext1.decrease_exponent_to(-10)
        encoded1 = phe.encoding.EncodedNumber.encode(self.public_key, 1)
        encoded2 = encoded1.decrease_exponent_to(-10)
        ciphertext = ciphertext1.decrease_exponent_to(-200)
        assert encoded2.exponent == -10
        assert ciphertext.exponent == -200
        ciphertext2 = ciphertext + encoded2

        self.assertEqual(16, self.private_key.decrypt(ciphertext2))

    def testMulWithEncryptedIntAndEncodedNumber(self):
        # Multiply two negative integers
        ciphertext1 = self.public_key.encrypt(-3)
        encoded2 = phe.encoding.EncodedNumber.encode(self.public_key, -25)
        ciphertext3 = ciphertext1 * encoded2
        decryption = self.private_key.decrypt(ciphertext3)
        self.assertEqual(75, decryption)

    def testEncryptFloatDecryptFloat4(self):
        # A small positive float
        ciphertext = self.public_key.encrypt(0.005743)
        decryption = self.private_key.decrypt(ciphertext)
        self.assertEqual(0.005743, decryption)

    def testEncryptFloatDecryptFloat5(self):
        # A small negative float
        ciphertext = self.public_key.encrypt(-0.05743)
        decryption = self.private_key.decrypt(ciphertext)
        self.assertEqual(-0.05743, decryption)

    def testAutomaticPrecision0(self):
        eps = sys.float_info.epsilon
        one_plus_eps = 1. + eps
        assert one_plus_eps > 1. # If this is false, we have trouble!

        ciphertext1 = self.public_key.encrypt(one_plus_eps)
        decryption1 = self.private_key.decrypt(ciphertext1)
        self.assertEqual(one_plus_eps, decryption1)

        ciphertext2 = ciphertext1 + eps
        self.assertGreater(ciphertext1.exponent, ciphertext2.exponent)
        decryption2 = self.private_key.decrypt(ciphertext2)
        self.assertEqual(one_plus_eps + eps, decryption2)

        # 1. + eps/5 == 1. for a python float...
        ciphertext3 = ciphertext1 + eps / 5
        decryption3 = self.private_key.decrypt(ciphertext3)
        self.assertEqual(one_plus_eps, decryption3)

        # ...but not for our 'arbitrary-precision' Paillier floats
        ciphertext4 = ciphertext3 + eps * 4. / 5
        decryption4 = self.private_key.decrypt(ciphertext4)
        self.assertNotEqual(one_plus_eps, decryption4)
        self.assertEqual(one_plus_eps + eps, decryption4)

    def testDecreaseExponentTo(self):
        # Decrease an exponent to -30 without affecting the plaintext number
        ciphertext1 = self.public_key.encrypt(1.01, precision=1e-8)
        assert -30 < ciphertext1.exponent # So the next part is meaningful
        ciphertext2 = ciphertext1.decrease_exponent_to(-30)

        self.assertLess(-30, ciphertext1.exponent)
        self.assertEqual(-30, ciphertext2.exponent)
        self.assertAlmostEqual(1.01, self.private_key.decrypt(ciphertext2),
                               places=8)

    def testDecreaseInvalidExponent(self):
        ciphertext = self.public_key.encrypt(1.01, precision=1e-8)
        assert ciphertext.exponent < 20
        self.assertRaises(ValueError, ciphertext.decrease_exponent_to, 20)


    def testAddWithEncryptDecryptFloat0(self):
        # Add 1 to a small negative number
        ciphertext1 = self.public_key.encrypt(-15.)
        ciphertext2 = self.public_key.encrypt(1.)
        ciphertext3 = ciphertext1 + ciphertext2
        decryption = self.private_key.decrypt(ciphertext3)
        self.assertEqual(-14., decryption)

    def testAddWithEncryptDecryptFloat0Right(self):
        # Add 1 to a small negative number
        ciphertext1 = self.public_key.encrypt(-15.)
        ciphertext2 = self.public_key.encrypt(1.)
        ciphertext3 = ciphertext2 + ciphertext1
        decryption = self.private_key.decrypt(ciphertext3)
        self.assertEqual(-14., decryption)

    def testAddWithEncryptDecryptFloat1(self):
        # Add 1 to a small positive number
        ciphertext1 = self.public_key.encrypt(15.)
        ciphertext2 = self.public_key.encrypt(1.)
        ciphertext3 = ciphertext1 + ciphertext2
        decryption = self.private_key.decrypt(ciphertext3)
        self.assertEqual(16., decryption)

    def testAddWithEncryptDecryptFloat2(self):
        # Add -1 to a small negative number
        ciphertext1 = self.public_key.encrypt(-15.)
        ciphertext2 = self.public_key.encrypt(-1.)
        ciphertext3 = ciphertext1 + ciphertext2
        decryption = self.private_key.decrypt(ciphertext3)
        self.assertEqual(-16., decryption)

    def testAddWithEncryptDecryptFloat3(self):
        # Add two floats with the same precision
        ciphertext1 = self.public_key.encrypt(1.3842)
        ciphertext2 = self.public_key.encrypt(-0.4)
        ciphertext3 = ciphertext1 + ciphertext2
        decryption = self.private_key.decrypt(ciphertext3)
        self.assertInRange(decryption, 0.9842, 1.3842 - 0.4)

    def testAddWithEncryptDecryptFloat4(self):
        # Add two floats with different precisions
        ciphertext1 = self.public_key.encrypt(0.1, precision=1e-3)
        ciphertext2 = self.public_key.encrypt(0.2, precision=1e-20)
        self.assertNotEqual(ciphertext1.exponent, ciphertext2.exponent)
        old_exponent = ciphertext1.exponent

        ciphertext3 = ciphertext1 + ciphertext2
        self.assertEqual(ciphertext2.exponent, ciphertext3.exponent)
        # Add should not have changed ciphertext1
        self.assertEqual(old_exponent, ciphertext1.exponent)

        decryption = self.private_key.decrypt(ciphertext3)
        self.assertAlmostEqual(0.3, decryption, places=3)

    def testSubWithEncryptDecryptFloat0(self):
        # Subtract two floats with different precisions
        ciphertext1 = self.public_key.encrypt(0.1, precision=1e-3)
        ciphertext2 = self.public_key.encrypt(0.2, precision=1e-20)
        self.assertNotEqual(ciphertext1.exponent, ciphertext2.exponent)

        ciphertext3 = ciphertext1 - ciphertext2
        self.assertEqual(ciphertext2.exponent, ciphertext3.exponent)

        decryption = self.private_key.decrypt(ciphertext3)
        self.assertAlmostEqual(-0.1, decryption, places=3) # Assumes base 10

    def testAddScalarWithEncryptDecryptFloat0(self):
        # Add a positive integer
        ciphertext1 = self.public_key.encrypt(-1.98)
        self.assertIsInstance(ciphertext1.ciphertext(False), int)
        ciphertext2 = ciphertext1 + 4

        self.assertEqual(2.02, self.private_key.decrypt(ciphertext2))

    def testAddScalarWithEncryptDecryptFloat0Right(self):
        # Add a positive integer
        ciphertext1 = self.public_key.encrypt(-1.98)
        self.assertIsInstance(ciphertext1.ciphertext(False), int)
        ciphertext2 = 4 + ciphertext1

        self.assertEqual(2.02, self.private_key.decrypt(ciphertext2))

    def testAddScalarWithEncryptDecryptFloat1(self):
        # Add a positive float
        ciphertext1 = self.public_key.encrypt(1.98)
        ciphertext2 = ciphertext1 + 4.3
        decryption = self.private_key.decrypt(ciphertext2)

        self.assertInRange(decryption, 1.98 + 4.3, 6.28)

    def testAddScalarWithEncryptDecryptFloat2(self):
        # Add a negative float
        ciphertext1 = self.public_key.encrypt(240.9)
        ciphertext2 = ciphertext1 +(- 40.8)
        decryption = self.private_key.decrypt(ciphertext2)

        self.assertInRange(decryption, 240.9 - 40.8, 200.1)

    def testAddScalarWithEncryptDecryptFloat3(self):
        # Add a negative integer
        ciphertext1 = self.public_key.encrypt(3.9)
        ciphertext2 = ciphertext1 +(- 40)

        self.assertEqual(-36.1, self.private_key.decrypt(ciphertext2))

    def testSubScalarWithEncryptDecryptFloat0(self):
        # Subtract a negative integer
        ciphertext1 = self.public_key.encrypt(-1.98)
        self.assertIsInstance(ciphertext1.ciphertext(False), int)
        ciphertext2 = ciphertext1 - (-4)

        self.assertEqual(2.02, self.private_key.decrypt(ciphertext2))

    def testSubScalarWithEncryptDecryptFloat0Right(self):
        # Subtract a negative integer
        ciphertext1 = self.public_key.encrypt(1.98)
        self.assertIsInstance(ciphertext1.ciphertext(False), int)
        ciphertext2 = 4 - ciphertext1

        self.assertEqual(2.02, self.private_key.decrypt(ciphertext2))

    def testSubScalarWithEncryptDecryptFloat1(self):
        # Subtract a negative float
        ciphertext1 = self.public_key.encrypt(1.98)
        ciphertext2 = ciphertext1 - (-4.3)

        decryption = self.private_key.decrypt(ciphertext2)
        self.assertInRange(decryption, 6.28, 1.98 -(-4.3))

    def testSubScalarWithEncryptDecryptFloat1Right(self):
        # Subtract a negative float
        ciphertext1 = self.public_key.encrypt(-1.98)
        ciphertext2 = (4.3) - ciphertext1

        decryption = self.private_key.decrypt(ciphertext2)
        self.assertInRange(decryption, 6.28, 1.98 -(-4.3))

    def testSubScalarWithEncryptDecryptFloat2(self):
        # Subtract a positive float
        ciphertext1 = self.public_key.encrypt(240.9)
        ciphertext2 = ciphertext1 - 40.8
        decryption = self.private_key.decrypt(ciphertext2)

        self.assertInRange(decryption, 200.1, 240.9 - 40.8)

    def testSubScalarWithEncryptDecryptFloat3(self):
        # Subtract a positive integer
        ciphertext1 = self.public_key.encrypt(3.9)
        ciphertext2 = ciphertext1 - 40

        self.assertEqual(-36.1, self.private_key.decrypt(ciphertext2))

    def testMulWithEncryptDecryptFloat0(self):
        # Multiply a floatish by one
        ciphertext1 = self.public_key.encrypt(-1.3)
        ciphertext2 = ciphertext1 * 1
        self.assertEqual(ciphertext1.exponent, ciphertext2.exponent)
        self.assertEqual(ciphertext1.ciphertext(False), ciphertext2.ciphertext(False))
        self.assertEqual(-1.3, self.private_key.decrypt(ciphertext2))

    def testMulWithEncryptDecryptFloat1(self):
        # Multiply a floatish by two
        ciphertext1 = self.public_key.encrypt(2.3)
        ciphertext2 = ciphertext1 * 2
        self.assertEqual(ciphertext1.exponent, ciphertext2.exponent)
        self.assertEqual(4.6, self.private_key.decrypt(ciphertext2))

    def testMulWithEncryptDecryptFloat2(self):
        # Multiply a floatish by a positive float
        ciphertext1 = self.public_key.encrypt(-0.1)
        ciphertext2 = ciphertext1 * 31.4
        self.assertEqual(-3.14, self.private_key.decrypt(ciphertext2))
        self.assertNotEqual(ciphertext2.exponent, ciphertext1.exponent)
        exp_of_314 = phe.encoding.EncodedNumber.encode(self.public_key, -31.4).exponent
        self.assertEqual(ciphertext2.exponent, ciphertext1.exponent +exp_of_314)

    def testMulWithEncryptedFloatAndEncodedNumber0(self):
        # Multiply a floatish with custom precision by a positive float
        ciphertext1 = self.public_key.encrypt(1.2345678e-12, precision=1e-14)
        encoded1 = phe.encoding.EncodedNumber.encode(self.public_key, 1.38734864,
                                                     precision=1e-2)
        ciphertext2 = ciphertext1 * encoded1
        self.assertAlmostEqual(1.71e-12, self.private_key.decrypt(ciphertext2), places=3)

    def testMulWithEncryptDecryptFloat4(self):
        # Multiply a floatish by minus one
        ciphertext1 = self.public_key.encrypt(-1.3)
        ciphertext2 = ciphertext1 * -1
        self.assertEqual(ciphertext1.exponent, ciphertext2.exponent)
        self.assertEqual(1.3, self.private_key.decrypt(ciphertext2))

    def testMulWithEncryptDecryptFloat5(self):
        # Multiply a floatish by minus two
        ciphertext1 = self.public_key.encrypt(2.3)
        ciphertext2 = ciphertext1 * -2
        self.assertEqual(ciphertext1.exponent, ciphertext2.exponent)
        self.assertEqual(-4.6, self.private_key.decrypt(ciphertext2))

    def testMulWithEncryptDecryptFloat6(self):
        # Multiply a floatish by a negative float
        ciphertext1 = self.public_key.encrypt(-0.1)
        ciphertext2 = ciphertext1 * -31.4
        self.assertEqual(3.14, self.private_key.decrypt(ciphertext2))
        self.assertNotEqual(ciphertext2.exponent, ciphertext1.exponent)
        exp_of_314 = phe.encoding.EncodedNumber.encode(self.public_key, -31.4).exponent
        self.assertEqual(ciphertext2.exponent, ciphertext1.exponent +exp_of_314)

    def testMulWithEncryptedFloatAndEncodedNumber1(self):
        # Multiply a floatish with custom precision by a negative float
        ciphertext1 = self.public_key.encrypt(1.2345678e-12, precision=1e-14)
        encoded1 = phe.encoding.EncodedNumber.encode(self.public_key, -1.38734864,
                                                     precision=1e-2)
        ciphertext2 = ciphertext1 * encoded1
        self.assertAlmostEqual(-1.71e-12, self.private_key.decrypt(ciphertext2), places=3)

    def testMulRight(self):
        # Check that it doesn't matter which side the real float is on
        ciphertext1 = self.public_key.encrypt(0.1)
        ciphertext2 = ciphertext1 * 31.4
        ciphertext3 = 31.4 * ciphertext1

        self.assertEqual(self.private_key.decrypt(ciphertext3), self.private_key.decrypt(ciphertext2))
        self.assertEqual(3.14, self.private_key.decrypt(ciphertext2))

    def testMulZero(self):
        # Check that multiplying by zero does something sensible
        ciphertext1 = self.public_key.encrypt(3.)
        ciphertext2 = ciphertext1 * 0
        self.assertEqual(0, self.private_key.decrypt(ciphertext2))

    def testMulZeroRight(self):
        # Check that multiplying by zero does something sensible
        ciphertext1 = self.public_key.encrypt(3.)
        ciphertext2 = 0 * ciphertext1
        self.assertEqual(0, self.private_key.decrypt(ciphertext2))

    def testDiv(self):
        # Check division works as well as multiplication does
        ciphertext1 = self.public_key.encrypt(6.28)
        ciphertext2 = ciphertext1 / 2
        self.assertEqual(3.14, self.private_key.decrypt(ciphertext2))

        ciphertext3 = ciphertext1 / 3.14
        self.assertEqual(2., self.private_key.decrypt(ciphertext3))

    def testAddWithEncryptedFloatAndEncodedNumber(self):
        # Add two floats with different precisions
        ciphertext1 = self.public_key.encrypt(0.1, precision=1e-3)
        encoded1 = phe.encoding.EncodedNumber.encode(self.public_key, 0.2,
                                                     precision=1e-20)
        self.assertNotEqual(ciphertext1.exponent, encoded1.exponent)
        old_exponent = ciphertext1.exponent

        ciphertext3 = ciphertext1 + encoded1
        self.assertEqual(encoded1.exponent, ciphertext3.exponent)
        # Add should not have changed ciphertext1
        self.assertEqual(old_exponent, ciphertext1.exponent)

        decryption = self.private_key.decrypt(ciphertext3)
        self.assertAlmostEqual(0.3, decryption, places=3)

    def testMulWithEncryptedFloatAndEncodedNumber(self):
        # Multiply a floatish by an encoded negative float
        ciphertext1 = self.public_key.encrypt(-0.1)
        encoded1 = phe.encoding.EncodedNumber.encode(self.public_key, -31.4)
        ciphertext2 = ciphertext1 * encoded1
        self.assertEqual(3.14, self.private_key.decrypt(ciphertext2))
        self.assertNotEqual(ciphertext2.exponent, ciphertext1.exponent)
        exp_of_314 = phe.encoding.EncodedNumber.encode(self.public_key, -31.4).exponent
        self.assertEqual(ciphertext2.exponent, ciphertext1.exponent +exp_of_314)

    def testObfuscate(self):
        ciphertext = self.public_key.encrypt(3.14)
        self.assertTrue(ciphertext._EncryptedNumber__is_obfuscated)
        c1 = ciphertext.ciphertext(False)
        ciphertext.obfuscate()
        self.assertTrue(ciphertext._EncryptedNumber__is_obfuscated)
        c2 = ciphertext.ciphertext(False)
        self.assertNotEqual(c1, c2)
        c3 = ciphertext.ciphertext(True)
        self.assertEqual(c2, c3)

    def testNotObfuscated(self):
        ciphertext = self.public_key.encrypt(3.14, r_value = 103)
        self.assertFalse(ciphertext._EncryptedNumber__is_obfuscated)
        c1 = ciphertext.ciphertext(be_secure=False)
        self.assertFalse(ciphertext._EncryptedNumber__is_obfuscated)
        c2 = ciphertext.ciphertext(be_secure=True)
        self.assertTrue(ciphertext._EncryptedNumber__is_obfuscated)
        c3 = ciphertext.ciphertext(be_secure=False)
        self.assertTrue(ciphertext._EncryptedNumber__is_obfuscated)
        c4 = ciphertext.ciphertext(be_secure=True)
        self.assertTrue(ciphertext._EncryptedNumber__is_obfuscated)

        self.assertNotEqual(c1, c2)
        self.assertEqual(c2, c3)
        self.assertEqual(c3, c4)
        dec = self.private_key.decrypt(ciphertext)
        self.assertEqual(3.14, dec)

    def testAddObfuscated(self):
        ciphertext1 = self.public_key.encrypt(94.5)
        ciphertext2 = self.public_key.encrypt(107.3)
        self.assertTrue(ciphertext1._EncryptedNumber__is_obfuscated)
        self.assertTrue(ciphertext2._EncryptedNumber__is_obfuscated)
        ciphertext3 = ciphertext1 + ciphertext2
        self.assertFalse(ciphertext3._EncryptedNumber__is_obfuscated)
        ciphertext3.ciphertext()
        self.assertTrue(ciphertext3._EncryptedNumber__is_obfuscated)

    def testEncryptInvalidType(self):
        data = "123"
        self.assertRaises(TypeError, self.public_key.encrypt, data)

    def testDecryptInvalidType(self):
        data = 123456789123456789123456789123456789
        ciphertext = self.public_key.encrypt(data)
        self.assertRaises(TypeError, self.private_key.decrypt, str(ciphertext))


class TestKeyring(unittest.TestCase):
    """Test adding and retrieving keys from a keyring."""
    def testKeyring(self):
        keyring1 = paillier.PaillierPrivateKeyring()
        public_key1, private_key1 = paillier.generate_paillier_keypair(keyring1)
        public_key2, private_key2 = paillier.generate_paillier_keypair(keyring1)
        self.assertEqual(private_key1, keyring1[public_key1])
        self.assertEqual(private_key2, keyring1[public_key2])

        ciphertext1 = public_key1.encrypt(5318008)
        ciphertext2 = public_key2.encrypt(1337)
        self.assertEqual(5318008, keyring1.decrypt(ciphertext1))
        self.assertEqual(1337,    keyring1.decrypt(ciphertext2))

        keyring2 = paillier.PaillierPrivateKeyring([private_key1, private_key2])
        self.assertEqual(keyring1, keyring2)
        self.assertRaises(TypeError, keyring1.add, public_key1)

        keyring1.add(private_key1)
        self.assertEqual(2, len(keyring1))

        del keyring1[public_key1]
        self.assertEqual(1, len(keyring1))

        self.assertRaises(KeyError, keyring1.decrypt, ciphertext1)


class TestIssue62(unittest.TestCase):
    def testIssue62(self):
        pub, priv = paillier.generate_paillier_keypair()
        a = pub.encrypt(445)
        # Force big exponent.
        b = pub.encrypt(0.16413409062205825) + pub.encrypt(2 ** -965)
        # This will raise OverflowError without bugfix #73.
        priv.decrypt(a + b)


def main():
    unittest.main()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# Portions copyright 2012 Google Inc. All Rights Reserved.
# This file has been modified by NICTA

# This file is part of pyphe.
#
# pyphe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyphe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyphe.  If not, see <http://www.gnu.org/licenses/>.

"""Paillier encryption library for partially homomorphic encryption."""
import random
import time
try:
    from collections.abc import Mapping
except ImportError:
    Mapping = dict

from phe import EncodedNumber
from phe.util import invert, powmod, mulmod, getprimeover, isqrt
import gpu_lib as gpu_cal
from test_gpu_lib import test_gxx
# Paillier cryptosystem is based on integer factorisation.
# The default is chosen to give a minimum of 128 bits of security.
# https://www.keylength.com/en/4/
# DEFAULT_KEYSIZE = 3072
DEFAULT_KEYSIZE = 1024



def generate_paillier_keypair(private_keyring=None, n_length=DEFAULT_KEYSIZE):
    """Return a new :class:`PaillierPublicKey` and :class:`PaillierPrivateKey`.

    Add the private key to *private_keyring* if given.

    Args:
      private_keyring (PaillierPrivateKeyring): a
        :class:`PaillierPrivateKeyring` on which to store the private
        key.
      n_length: key size in bits.

    Returns:
      tuple: The generated :class:`PaillierPublicKey` and
      :class:`PaillierPrivateKey`
    """
    p = q = n = None
    n_len = 0
    while n_len != n_length:
        p = getprimeover(n_length // 2)
        q = p
        while q == p:
            q = getprimeover(n_length // 2)
        n = p * q
        n_len = n.bit_length()

    public_key = PaillierPublicKey(n)
    private_key = PaillierPrivateKey(public_key, p, q)

    if private_keyring is not None:
        private_keyring.add(private_key)

    return public_key, private_key


class PaillierPublicKey(object):
    """Contains a public key and associated encryption methods.

    Args:

      n (int): the modulus of the public key - see Paillier's paper.

    Attributes:
      g (int): part of the public key - see Paillier's paper.
      n (int): part of the public key - see Paillier's paper.
      nsquare (int): :attr:`n` ** 2, stored for frequent use.
      max_int (int): Maximum int that may safely be stored. This can be
        increased, if you are happy to redefine "safely" and lower the
        chance of detecting an integer overflow.
    """
    def __init__(self, n):
        self.g = n + 1
        self.n = n
        self.nsquare = n * n
        self.max_int = n // 3 - 1

    def __repr__(self):
        publicKeyHash = hex(hash(self))[2:]
        return "<PaillierPublicKey {}>".format(publicKeyHash[:10])

    def __eq__(self, other):
        return self.n == other.n

    def __hash__(self):
        return hash(self.n)
    def gpu_raw_encrypt(self, plaintextlist, r_value=None):
        batch_param = []
        batch_param1 = []
        batch_param2 = []
        nude_ciphertext_list = []
        start = time.time()
        for plaintext in plaintextlist:
            if not isinstance(plaintext, int):
                raise TypeError('Expected int type plaintext but got: %s' %
                                type(plaintext))

            if self.n - self.max_int <= plaintext < self.n:
                # Very large plaintext, take a sneaky shortcut using inverses
                neg_plaintext = self.n - plaintext  # = abs(plaintext - nsquare)
                # avoid using gmpy2's mulmod when a * b < c
                neg_ciphertext = (self.n * neg_plaintext + 1) % self.nsquare
                nude_ciphertext = invert(neg_ciphertext, self.nsquare)
                nude_ciphertext_list.append(nude_ciphertext)
            else:
                # we chose g = n + 1, so that we can exploit the fact that
                # (n+1)^plaintext = n*plaintext + 1 mod n^2
                nude_ciphertext = (self.n * plaintext + 1) % self.nsquare
                nude_ciphertext_list.append(nude_ciphertext)
        r = r_value or self.get_random_lt_n()
        obfuscator = powmod(r, self.n, self.nsquare)
        r_bytes = obfuscator.to_bytes(512, "little")
        nsquare_bytes = self.nsquare.to_bytes(512, "little")
        for i in nude_ciphertext_list:
            nude_ciphertext_bytes = i.to_bytes(512, "little")
            batch_param.append((nude_ciphertext_bytes,r_bytes,nsquare_bytes ))
        result= gpu_cal.mulm_2048(batch_param, len(batch_param))
        result2 = [int.from_bytes(i, "little")for i in result]
        return result2
        
    def raw_encrypt(self, plaintext, r_value=None):
        """Paillier encryption of a positive integer plaintext < :attr:`n`.

        You probably should be using :meth:`encrypt` instead, because it
        handles positive and negative ints and floats.

        Args:
          plaintext (int): a positive integer < :attr:`n` to be Paillier
            encrypted. Typically this is an encoding of the actual
            number you want to encrypt.
          r_value (int): obfuscator for the ciphertext; by default (i.e.
            r_value is None), a random value is used.

        Returns:
          int: Paillier encryption of plaintext.

        Raises:
          TypeError: if plaintext is not an int.
        """
        if not isinstance(plaintext, int):
            raise TypeError('Expected int type plaintext but got: %s' %
                            type(plaintext))

        if self.n - self.max_int <= plaintext < self.n:
            # Very large plaintext, take a sneaky shortcut using inverses
            neg_plaintext = self.n - plaintext  # = abs(plaintext - nsquare)
            # avoid using gmpy2's mulmod when a * b < c
            neg_ciphertext = (self.n * neg_plaintext + 1) % self.nsquare
            nude_ciphertext = invert(neg_ciphertext, self.nsquare)
        else:
            # we chose g = n + 1, so that we can exploit the fact that
            # (n+1)^plaintext = n*plaintext + 1 mod n^2
            nude_ciphertext = (self.n * plaintext + 1) % self.nsquare

        r = r_value or self.get_random_lt_n()
        obfuscator = powmod(r, self.n, self.nsquare)

        return mulmod(nude_ciphertext, obfuscator, self.nsquare)

    def get_random_lt_n(self):
        """Return a cryptographically random number less than :attr:`n`"""
        return random.SystemRandom().randrange(1, self.n)
    def gpu_encrypt(self, list, precision=None, r_value=None):
      encode_list = []
      for value in list:
        if isinstance(value, EncodedNumber):
            encoding = value
            encode_list.append(encoding)
        else:
            encoding = EncodedNumber.encode(self, value, precision)
            encode_list.append(encoding)
      return self.gpu_encrypt_encoded(encode_list, r_value)

    def encrypt(self, value, precision=None, r_value=None):
        """Encode and Paillier encrypt a real number *value*.

        Args:
          value: an int or float to be encrypted.
            If int, it must satisfy abs(*value*) < :attr:`n`/3.
            If float, it must satisfy abs(*value* / *precision*) <<
            :attr:`n`/3
            (i.e. if a float is near the limit then detectable
            overflow may still occur)
          precision (float): Passed to :meth:`EncodedNumber.encode`.
            If *value* is a float then *precision* is the maximum
            **absolute** error allowed when encoding *value*. Defaults
            to encoding *value* exactly.
          r_value (int): obfuscator for the ciphertext; by default (i.e.
            if *r_value* is None), a random value is used.

        Returns:
          EncryptedNumber: An encryption of *value*.

        Raises:
          ValueError: if *value* is out of range or *precision* is so
            high that *value* is rounded to zero.
        """

        if isinstance(value, EncodedNumber):
            encoding = value
        else:
            encoding = EncodedNumber.encode(self, value, precision)

        return self.encrypt_encoded(encoding, r_value)

    def encrypt_encoded(self, encoding, r_value):
        """Paillier encrypt an encoded value.

        Args:
          encoding: The EncodedNumber instance.
          r_value (int): obfuscator for the ciphertext; by default (i.e.
            if *r_value* is None), a random value is used.

        Returns:
          EncryptedNumber: An encryption of *value*.
        """
        # If r_value is None, obfuscate in a call to .obfuscate() (below)
        obfuscator = r_value or 1
        ciphertext = self.raw_encrypt(encoding.encoding, r_value=obfuscator)
        encrypted_number = EncryptedNumber(self, ciphertext, encoding.exponent)
        if r_value is None:
            encrypted_number.obfuscate()
        return encrypted_number

    def gpu_encrypt_encoded(self, encoding_list, r_value):
        """Paillier encrypt an encoded value.

        Args:
          encoding: The EncodedNumber instance.
          r_value (int): obfuscator for the ciphertext; by default (i.e.
            if *r_value* is None), a random value is used.

        Returns:
          EncryptedNumber: An encryption of *value*.
        """
        # If r_value is None, obfuscate in a call to .obfuscate() (below)
        obfuscator = r_value or 1
        instance_list = [i.encoding for i in encoding_list]
        ciphertext_list = self.gpu_raw_encrypt(instance_list, r_value=obfuscator)
        encrypted_number_list = [EncryptedNumber(self, ciphertext, encoding.exponent)for ciphertext,encoding in zip(ciphertext_list,encoding_list)]
        # if r_value is None:
        #     encrypted_number.obfuscate()
        return encrypted_number_list



class PaillierPrivateKey(object):
    """Contains a private key and associated decryption method.

    Args:
      public_key (:class:`PaillierPublicKey`): The corresponding public
        key.
      p (int): private secret - see Paillier's paper.
      q (int): private secret - see Paillier's paper.

    Attributes:
      public_key (PaillierPublicKey): The corresponding public
        key.
      p (int): private secret - see Paillier's paper.
      q (int): private secret - see Paillier's paper.
      psquare (int): p^2
      qsquare (int): q^2
      p_inverse (int): p^-1 mod q
      hp (int): h(p) - see Paillier's paper.
      hq (int): h(q) - see Paillier's paper.
    """
    def __init__(self, public_key, p, q):
        if not p*q == public_key.n:
            raise ValueError('given public key does not match the given p and q.')
        if p == q:
            # check that p and q are different, otherwise we can't compute p^-1 mod q
            raise ValueError('p and q have to be different')
        self.public_key = public_key
        if q < p: #ensure that p < q.
            self.p = q
            self.q = p
        else:
            self.p = p
            self.q = q
        self.psquare = self.p * self.p
        self.qsquare = self.q * self.q
        self.lam = (self.p-1) * (self.q-1)
        self.mu = invert(self.l_function(powmod(self.public_key.g,self.lam,self.public_key.nsquare),self.public_key.n),self.public_key.n)
        self.p_inverse = invert(self.p, self.q)
        self.hp = self.h_function(self.p, self.psquare)
        self.hq = self.h_function(self.q, self.qsquare)

    @staticmethod
    def from_totient(public_key, totient):
        """given the totient, one can factorize the modulus

        The totient is defined as totient = (p - 1) * (q - 1),
        and the modulus is defined as modulus = p * q

        Args:
          public_key (PaillierPublicKey): The corresponding public
            key
          totient (int): the totient of the modulus

        Returns:
          the :class:`PaillierPrivateKey` that corresponds to the inputs

        Raises:
          ValueError: if the given totient is not the totient of the modulus
            of the given public key
        """
        p_plus_q = public_key.n - totient + 1
        p_minus_q = isqrt(p_plus_q * p_plus_q - public_key.n * 4)
        q = (p_plus_q - p_minus_q) // 2
        p = p_plus_q - q
        if not p*q == public_key.n:
            raise ValueError('given public key and totient do not match.')
        return PaillierPrivateKey(public_key, p, q)

    def __repr__(self):
        pub_repr = repr(self.public_key)
        return "<PaillierPrivateKey for {}>".format(pub_repr)
        
    def gpu_decrypt(self, encrypted_number_list):
        
        encoded_list = self.gpu_decrypt_encoded(encrypted_number_list)
        return [encoded.decode()for encoded in encoded_list]

    def decrypt(self, encrypted_number):
        """Return the decrypted & decoded plaintext of *encrypted_number*.

        Uses the default :class:`EncodedNumber`, if using an alternative encoding
        scheme, use :meth:`decrypt_encoded` or :meth:`raw_decrypt` instead.

        Args:
          encrypted_number (EncryptedNumber): an
            :class:`EncryptedNumber` with a public key that matches this
            private key.

        Returns:
          the int or float that `EncryptedNumber` was holding. N.B. if
            the number returned is an integer, it will not be of type
            float.

        Raises:
          TypeError: If *encrypted_number* is not an
            :class:`EncryptedNumber`.
          ValueError: If *encrypted_number* was encrypted against a
            different key.
        """
        encoded = self.decrypt_encoded(encrypted_number)
        return encoded.decode()
    
    def gpu_decrypt_encoded(self, encrypted_number_list, Encoding=None):
        encoded_list = []
        for encrypted_number in encrypted_number_list:
            if not isinstance(encrypted_number, EncryptedNumber):
                raise TypeError('Expected encrypted_number to be an EncryptedNumber'
                                ' not: %s' % type(encrypted_number))
            if self.public_key != encrypted_number.public_key:
                raise ValueError('encrypted_number was encrypted against a '
                                'different key!')
        if Encoding is None:
            Encoding = EncodedNumber
        res_list = [encrypted_number.ciphertext(be_secure=False)for encrypted_number in encrypted_number_list]
        encoded_list = self.gpu_raw_decrypt(res_list)
        return [Encoding(self.public_key, encoded,
                             encrypted_number.exponent)for encoded,encrypted_number in zip(encoded_list,encrypted_number_list) ]

    def decrypt_encoded(self, encrypted_number, Encoding=None):
        """Return the :class:`EncodedNumber` decrypted from *encrypted_number*.

        Args:
          encrypted_number (EncryptedNumber): an
            :class:`EncryptedNumber` with a public key that matches this
            private key.
          Encoding (class): A class to use instead of :class:`EncodedNumber`, the
            encoding used for the *encrypted_number* - used to support alternative
            encodings.

        Returns:
          :class:`EncodedNumber`: The decrypted plaintext.

        Raises:
          TypeError: If *encrypted_number* is not an
            :class:`EncryptedNumber`.
          ValueError: If *encrypted_number* was encrypted against a
            different key.
        """
        if not isinstance(encrypted_number, EncryptedNumber):
            raise TypeError('Expected encrypted_number to be an EncryptedNumber'
                            ' not: %s' % type(encrypted_number))

        if self.public_key != encrypted_number.public_key:
            raise ValueError('encrypted_number was encrypted against a '
                             'different key!')

        if Encoding is None:
            Encoding = EncodedNumber

        encoded = self.raw_decrypt(encrypted_number.ciphertext(be_secure=False))
        return Encoding(self.public_key, encoded,
                             encrypted_number.exponent)

    def gpu_raw_decrypt(self, ciphertext_list):
        for ciphertext in ciphertext_list:
            if not isinstance(ciphertext, int):
                raise TypeError('Expected ciphertext to be an int, not: %s' %
                    type(ciphertext))
        batch_param = []
        batch_param1 = []
        batch_param2 = []
        batch_param3 = []
        p_1_bytes = (self.p-1).to_bytes(512, "little")
        psquare_bytes = self.psquare.to_bytes(512, "little")
        q_1_bytes = (self.q-1).to_bytes(512, "little")
        qsquare_bytes = self.qsquare.to_bytes(512, "little")
        hp_bytes = self.hp.to_bytes(512, "little")
        hq_bytes = self.hq.to_bytes(512, "little")
        p_bytes = self.p.to_bytes(512, "little")
        q_bytes = self.q.to_bytes(512, "little")
        for ciphertext in ciphertext_list:
            t1 = ciphertext
            t2 = ciphertext
            if ciphertext > self.psquare :
                t1 = ciphertext %  self.psquare
            if ciphertext > self.qsquare :
                t2 = ciphertext %  self.qsquare
            ciphertext_bytes_p = t1.to_bytes(512, "little")
            ciphertext_bytes_q = t2.to_bytes(512, "little")
            batch_param.append((ciphertext_bytes_p, p_1_bytes, psquare_bytes))
            # GPU 调用异常，当ciphertext_bytes<  psquare_bytes CGNB模乘不调用 显示异常
            batch_param1.append((ciphertext_bytes_q, q_1_bytes, qsquare_bytes))
        result1 = gpu_cal.powm_2048(batch_param, len(batch_param))
        result2 = gpu_cal.powm_2048(batch_param1, len(batch_param1))
        result3 = [self.l_function(int.from_bytes(i, "little"), self.p)for i in result1]
        result4 = [self.l_function(int.from_bytes(i, "little"), self.q)for i in result2]
        for j,k in zip(result3,result4):
            j_bytes = j.to_bytes(512, "little")
            k_bytes = k.to_bytes(512, "little")
            batch_param2.append((j_bytes, hp_bytes, p_bytes))
            batch_param3.append((k_bytes, hq_bytes, q_bytes))
        result5 = gpu_cal.mulm_2048(batch_param2, len(batch_param2))
        result6 = gpu_cal.mulm_2048(batch_param3, len(batch_param3))
        result7 = [int.from_bytes(i, "little")for i in result5]
        result8 = [int.from_bytes(i, "little")for i in result6]
        return self.gpu_crt(result7,result8)

    def raw_decrypt(self, ciphertext):
        """Decrypt raw ciphertext and return raw plaintext.

        Args:
          ciphertext (int): (usually from :meth:`EncryptedNumber.ciphertext()`)
            that is to be Paillier decrypted.

        Returns:
          int: Paillier decryption of ciphertext. This is a positive
          integer < :attr:`public_key.n`.

        Raises:
          TypeError: if ciphertext is not an int.
        """
        if not isinstance(ciphertext, int):
            raise TypeError('Expected ciphertext to be an int, not: %s' %
                type(ciphertext))

        decrypt_to_p = mulmod(
            self.l_function(powmod(ciphertext, self.p-1, self.psquare), self.p),
            self.hp,
            self.p)
        t = powmod(ciphertext, self.q-1, self.qsquare)
        decrypt_to_q = mulmod(
            self.l_function(t, self.q),
            self.hq,
            self.q)
        return self.crt(decrypt_to_p, decrypt_to_q)

    def h_function(self, x, xsquare):
        """Computes the h-function as defined in Paillier's paper page 12,
        'Decryption using Chinese-remaindering'.
        """
        return invert(self.l_function(powmod(self.public_key.g, x - 1, xsquare),x), x)

    def l_function(self, x, p):
        """Computes the L function as defined in Paillier's paper. That is: L(x,p) = (x-1)/p"""
        return (x - 1) // p
    
    def gpu_crt(self, mp_list, mq_list):
        batch_param = []
        p_inverse_bytes = self.p_inverse.to_bytes(512, "little")
        q_bytes = self.q.to_bytes(512, "little")
        for mp,mq in zip(mp_list, mq_list):
            mq_mp_bytes = (mq - mp).to_bytes(512, "little")
            batch_param.append((mq_mp_bytes, p_inverse_bytes, q_bytes))
        result = gpu_cal.mulm_2048(batch_param, len(batch_param))
        return [mp + (int.from_bytes(u,"little") * self.p)for mp,u in zip(mp_list,result)]

    def crt(self, mp, mq):
        """The Chinese Remainder Theorem as needed for decryption. Returns the solution modulo n=pq.

        Args:
           mp(int): the solution modulo p.
           mq(int): the solution modulo q.
       """
        u = mulmod(mq - mp, self.p_inverse, self.q)
        return mp + (u * self.p)

    def __eq__(self, other):
        return self.p == other.p and self.q == other.q

    def __hash__(self):
        return hash((self.p, self.q))


class PaillierPrivateKeyring(Mapping):
    """Holds several private keys and can decrypt using any of them.

    Acts like a dict, supports :func:`del`, and indexing with **[]**,
    but adding keys is done using :meth:`add`.

    Args:
      private_keys (list of PaillierPrivateKey): an optional starting
        list of :class:`PaillierPrivateKey` instances.
    """
    def __init__(self, private_keys=None):
        if private_keys is None:
            private_keys = []
        public_keys = [k.public_key for k in private_keys]
        self.__keyring = dict(zip(public_keys, private_keys))

    def __getitem__(self, key):
        return self.__keyring[key]

    def __len__(self):
        return len(self.__keyring)

    def __iter__(self):
        return iter(self.__keyring)

    def __delitem__(self, public_key):
        del self.__keyring[public_key]

    def add(self, private_key):
        """Add a key to the keyring.

        Args:
          private_key (PaillierPrivateKey): a key to add to this keyring.
        """
        if not isinstance(private_key, PaillierPrivateKey):
            raise TypeError("private_key should be of type PaillierPrivateKey, "
                            "not %s" % type(private_key))
        self.__keyring[private_key.public_key] = private_key
    def gpu_encrypt(self,encrypted_number_list):
        relevant_private_key = self.__keyring[encrypted_number.public_key]
        return relevant_private_key.gpu_decrypt(encrypted_number_list)
    def decrypt(self, encrypted_number):
        """Return the decrypted & decoded plaintext of *encrypted_number*.

        Args:
          encrypted_number (EncryptedNumber): encrypted against a known public
            key, i.e., one for which the private key is on this keyring.

        Returns:
          the int or float that *encrypted_number* was holding. N.B. if
          the number returned is an integer, it will not be of type
          float.

        Raises:
          KeyError: If the keyring does not hold the private key that
            decrypts *encrypted_number*.
        """
        relevant_private_key = self.__keyring[encrypted_number.public_key]
        return relevant_private_key.decrypt(encrypted_number)


class EncryptedNumber(object):
    """Represents the Paillier encryption of a float or int.

    Typically, an `EncryptedNumber` is created by
    :meth:`PaillierPublicKey.encrypt`. You would only instantiate an
    `EncryptedNumber` manually if you are de-serializing a number
    someone else encrypted.


    Paillier encryption is only defined for non-negative integers less
    than :attr:`PaillierPublicKey.n`. :class:`EncodedNumber` provides
    an encoding scheme for floating point and signed integers that is
    compatible with the partially homomorphic properties of the Paillier
    cryptosystem:

    1. D(E(a) * E(b)) = a + b
    2. D(E(a)**b)     = a * b

    where `a` and `b` are ints or floats, `E` represents encoding then
    encryption, and `D` represents decryption then decoding.

    Args:
      public_key (PaillierPublicKey): the :class:`PaillierPublicKey`
        against which the number was encrypted.
      ciphertext (int): encrypted representation of the encoded number.
      exponent (int): used by :class:`EncodedNumber` to keep track of
        fixed precision. Usually negative.

    Attributes:
      public_key (PaillierPublicKey): the :class:`PaillierPublicKey`
        against which the number was encrypted.
      exponent (int): used by :class:`EncodedNumber` to keep track of
        fixed precision. Usually negative.

    Raises:
      TypeError: if *ciphertext* is not an int, or if *public_key* is
        not a :class:`PaillierPublicKey`.
    """
    def __init__(self, public_key, ciphertext, exponent=0):
        self.public_key = public_key
        self.__ciphertext = ciphertext
        self.exponent = exponent
        self.__is_obfuscated = False
        if isinstance(self.ciphertext, EncryptedNumber):
            raise TypeError('ciphertext should be an integer')
        if not isinstance(self.public_key, PaillierPublicKey):
            raise TypeError('public_key should be a PaillierPublicKey')

    def __add__(self, other):
        """Add an int, float, `EncryptedNumber` or `EncodedNumber`."""
        if isinstance(other, EncryptedNumber):
            return self._add_encrypted(other)
        elif isinstance(other, EncodedNumber):
            return self._add_encoded(other)
        else:
            return self._add_scalar(other)

    def __radd__(self, other):
        """Called when Python evaluates `34 + <EncryptedNumber>`
        Required for builtin `sum` to work.
        """
        return self.__add__(other)

    def __mul__(self, other):
        """Multiply by an int, float, or EncodedNumber."""
        if isinstance(other, EncryptedNumber):
            raise NotImplementedError('Good luck with that...')

        if isinstance(other, EncodedNumber):
            encoding = other
        else:
            encoding = EncodedNumber.encode(self.public_key, other)
        product = self._raw_mul(encoding.encoding)
        exponent = self.exponent + encoding.exponent

        return EncryptedNumber(self.public_key, product, exponent)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __sub__(self, other):
        return self + (other * -1)

    def __rsub__(self, other):
        return other + (self * -1)

    def __truediv__(self, scalar):
        return self.__mul__(1 / scalar)

    def ciphertext(self, be_secure=True):
        """Return the ciphertext of the EncryptedNumber.

        Choosing a random number is slow. Therefore, methods like
        :meth:`__add__` and :meth:`__mul__` take a shortcut and do not
        follow Paillier encryption fully - every encrypted sum or
        product should be multiplied by r **
        :attr:`~PaillierPublicKey.n` for random r < n (i.e., the result
        is obfuscated). Not obfuscating provides a big speed up in,
        e.g., an encrypted dot product: each of the product terms need
        not be obfuscated, since only the final sum is shared with
        others - only this final sum needs to be obfuscated.

        Not obfuscating is OK for internal use, where you are happy for
        your own computer to know the scalars you've been adding and
        multiplying to the original ciphertext. But this is *not* OK if
        you're going to be sharing the new ciphertext with anyone else.

        So, by default, this method returns an obfuscated ciphertext -
        obfuscating it if necessary. If instead you set `be_secure=False`
        then the ciphertext will be returned, regardless of whether it
        has already been obfuscated. We thought that this approach,
        while a little awkward, yields a safe default while preserving
        the option for high performance.

        Args:
          be_secure (bool): If any untrusted parties will see the
            returned ciphertext, then this should be True.

        Returns:
          an int, the ciphertext. If `be_secure=False` then it might be
            possible for attackers to deduce numbers involved in
            calculating this ciphertext.
        """
        if be_secure and not self.__is_obfuscated:
            self.obfuscate()

        return self.__ciphertext

    def decrease_exponent_to(self, new_exp):
        """Return an EncryptedNumber with same value but lower exponent.

        If we multiply the encoded value by :attr:`EncodedNumber.BASE` and
        decrement :attr:`exponent`, then the decoded value does not change.
        Thus we can almost arbitrarily ratchet down the exponent of an
        `EncryptedNumber` - we only run into trouble when the encoded
        integer overflows. There may not be a warning if this happens.

        When adding `EncryptedNumber` instances, their exponents must
        match.

        This method is also useful for hiding information about the
        precision of numbers - e.g. a protocol can fix the exponent of
        all transmitted `EncryptedNumber` instances to some lower bound(s).

        Args:
          new_exp (int): the desired exponent.

        Returns:
          EncryptedNumber: Instance with the same plaintext and desired
            exponent.

        Raises:
          ValueError: You tried to increase the exponent.
        """
        if new_exp > self.exponent:
            raise ValueError('New exponent %i should be more negative than '
                             'old exponent %i' % (new_exp, self.exponent))
        multiplied = self * pow(EncodedNumber.BASE, self.exponent - new_exp)
        multiplied.exponent = new_exp
        return multiplied

    def obfuscate(self):
        """Disguise ciphertext by multiplying by r ** n with random r.

        This operation must be performed for every `EncryptedNumber`
        that is sent to an untrusted party, otherwise eavesdroppers
        might deduce relationships between this and an antecedent
        `EncryptedNumber`.

        For example::

            enc = public_key.encrypt(1337)
            send_to_nsa(enc)       # NSA can't decrypt (we hope!)
            product = enc * 3.14
            send_to_nsa(product)   # NSA can deduce 3.14 by bruteforce attack
            product2 = enc * 2.718
            product2.obfuscate()
            send_to_nsa(product)   # NSA can't deduce 2.718 by bruteforce attack
        """
        r = self.public_key.get_random_lt_n()
        r_pow_n = powmod(r, self.public_key.n, self.public_key.nsquare)
        self.__ciphertext = mulmod(self.__ciphertext, r_pow_n, self.public_key.nsquare)
        self.__is_obfuscated = True

    def _add_scalar(self, scalar):
        """Returns E(a + b), given self=E(a) and b.

        Args:
          scalar: an int or float b, to be added to `self`.

        Returns:
          EncryptedNumber: E(a + b), calculated by encrypting b and
            taking the product of E(a) and E(b) modulo
            :attr:`~PaillierPublicKey.n` ** 2.

        Raises:
          ValueError: if scalar is out of range or precision.
        """
        encoded = EncodedNumber.encode(self.public_key, scalar,
                                       max_exponent=self.exponent)

        return self._add_encoded(encoded)

    def _add_encoded(self, encoded):
        """Returns E(a + b), given self=E(a) and b.

        Args:
          encoded (EncodedNumber): an :class:`EncodedNumber` to be added
            to `self`.

        Returns:
          EncryptedNumber: E(a + b), calculated by encrypting b and
            taking the product of E(a) and E(b) modulo
            :attr:`~PaillierPublicKey.n` ** 2.

        Raises:
          ValueError: if scalar is out of range or precision.
        """
        if self.public_key != encoded.public_key:
            raise ValueError("Attempted to add numbers encoded against "
                             "different public keys!")

        # In order to add two numbers, their exponents must match.
        a, b = self, encoded
        if a.exponent > b.exponent:
            a = self.decrease_exponent_to(b.exponent)
        elif a.exponent < b.exponent:
            b = b.decrease_exponent_to(a.exponent)

        # Don't bother to salt/obfuscate in a basic operation, do it
        # just before leaving the computer.
        encrypted_scalar = a.public_key.raw_encrypt(b.encoding, 1)

        sum_ciphertext = a._raw_add(a.ciphertext(False), encrypted_scalar)
        return EncryptedNumber(a.public_key, sum_ciphertext, a.exponent)

    def _add_encrypted(self, other):
        """Returns E(a + b) given E(a) and E(b).

        Args:
          other (EncryptedNumber): an `EncryptedNumber` to add to self.

        Returns:
          EncryptedNumber: E(a + b), calculated by taking the product
            of E(a) and E(b) modulo :attr:`~PaillierPublicKey.n` ** 2.

        Raises:
          ValueError: if numbers were encrypted against different keys.
        """
        if self.public_key != other.public_key:
            raise ValueError("Attempted to add numbers encrypted against "
                             "different public keys!")

        # In order to add two numbers, their exponents must match.
        a, b = self, other
        if a.exponent > b.exponent:
            a = self.decrease_exponent_to(b.exponent)
        elif a.exponent < b.exponent:
            b = b.decrease_exponent_to(a.exponent)

        sum_ciphertext = a._raw_add(a.ciphertext(False), b.ciphertext(False))
        return EncryptedNumber(a.public_key, sum_ciphertext, a.exponent)

    def _raw_add(self, e_a, e_b):
        """Returns the integer E(a + b) given ints E(a) and E(b).

        N.B. this returns an int, not an `EncryptedNumber`, and ignores
        :attr:`ciphertext`

        Args:
          e_a (int): E(a), first term
          e_b (int): E(b), second term

        Returns:
          int: E(a + b), calculated by taking the product of E(a) and
            E(b) modulo :attr:`~PaillierPublicKey.n` ** 2.
        """
        return mulmod(e_a, e_b, self.public_key.nsquare)

    def _raw_mul(self, plaintext):
        """Returns the integer E(a * plaintext), where E(a) = ciphertext

        Args:
          plaintext (int): number by which to multiply the
            `EncryptedNumber`. *plaintext* is typically an encoding.
            0 <= *plaintext* < :attr:`~PaillierPublicKey.n`

        Returns:
          int: Encryption of the product of `self` and the scalar
            encoded in *plaintext*.

        Raises:
          TypeError: if *plaintext* is not an int.
          ValueError: if *plaintext* is not between 0 and
            :attr:`PaillierPublicKey.n`.
        """
        if not isinstance(plaintext, int):
            raise TypeError('Expected ciphertext to be int, not %s' %
                type(plaintext))

        if plaintext < 0 or plaintext >= self.public_key.n:
            raise ValueError('Scalar out of bounds: %i' % plaintext)

        if self.public_key.n - self.public_key.max_int <= plaintext:
            # Very large plaintext, play a sneaky trick using inverses
            neg_c = invert(self.ciphertext(False), self.public_key.nsquare)
            neg_scalar = self.public_key.n - plaintext
            return powmod(neg_c, neg_scalar, self.public_key.nsquare)
        else:
            return powmod(self.ciphertext(False), plaintext, self.public_key.nsquare)



import json
import random
from unittest import TestCase
import tempfile

from click.testing import CliRunner

from phe.command_line import cli


class TestConsoleBasics(TestCase):

    def test_cli_includes_help(self):
        runner = CliRunner()
        result = runner.invoke(cli, ['--help'])
        assert result.exit_code == 0

        assert 'Usage' in result.output
        assert 'Options' in result.output
        assert 'Commands' in result.output

    def test_generate_keypair_to_file(self):
        runner = CliRunner()

        with tempfile.NamedTemporaryFile() as outfile:
            result = runner.invoke(cli, ['genpkey', '--keysize', '256', outfile.name])
            print(result.output)
            assert result.exit_code == 0

            outfile.seek(0)
            written_data = outfile.read()

            priv_key = json.loads(written_data.decode('utf-8'))

        assert 'pub' in priv_key
        assert 'kty' in priv_key
        assert 'p' in priv_key

    def test_generate_keypair_to_stdout(self):
        runner = CliRunner()

        result = runner.invoke(cli, ['genpkey', '--keysize', '256', '-'])

        assert 'pub' in result.output
        assert 'kty' in result.output
        assert 'p' in result.output

    def test_extract_public_key(self):
        runner = CliRunner()

        with tempfile.NamedTemporaryFile() as private_keyfile:
            runner.invoke(cli, ['genpkey', '--keysize', '256', private_keyfile.name])

            with tempfile.NamedTemporaryFile() as public_keyfile:
                result = runner.invoke(cli, ['extract', private_keyfile.name, public_keyfile.name])
                assert result.exit_code == 0

                public_keyfile.seek(0)
                written_data = public_keyfile.read().decode('utf-8')

                assert '"kty":' in written_data
                assert '"n":' in written_data
                assert '"alg":' in written_data

                assert '"p":' not in written_data
                assert '"q":' not in written_data


class TestConsoleEncryption(TestCase):

    @classmethod
    def setUpClass(cls):
        """Generate a keypair and extract the public key.
        """
        cls.private_keyfile = tempfile.NamedTemporaryFile()
        cls.public_keyfile = tempfile.NamedTemporaryFile()

        cls.runner = CliRunner()
        cls.runner.invoke(cli, ['genpkey', '--keysize', '256', cls.private_keyfile.name])
        cls.runner.invoke(cli, ['extract', cls.private_keyfile.name, cls.public_keyfile.name])

    @classmethod
    def tearDownClass(cls):
        cls.private_keyfile.close()
        cls.public_keyfile.close()

    def test_encrypt_positive_integers(self):
        numbers = [0, 1, 2, 5, 10, '1', '10550']

        for num in numbers:
            result = self.runner.invoke(cli, ['encrypt', self.public_keyfile.name, str(num)])
            assert result.exit_code == 0

    def test_encrypt_signed_integers(self):
        """encrypting positive and negative integer"""
        numbers = [0, 1, -1, 10, '1', '-10550']

        for num in numbers:
            result = self.runner.invoke(cli, ['encrypt', self.public_keyfile.name, "--", str(num)])
            assert result.exit_code == 0

    def test_encrypt_float(self):
        numbers = [0.0, 1.1, -0.0001, 100000.01, '1e-20', '-10550e20']

        for num in numbers:
            result = self.runner.invoke(cli, ['encrypt', self.public_keyfile.name, "--", str(num)])
            assert result.exit_code == 0

    def test_encrypt_to_stdout(self):
        """Test encrypting and writing output to a file"""
        numbers = [0.0, 1.1, -0.0001, 100000.01, '1e-20', '-10550e20']

        for num in numbers:
            result = self.runner.invoke(cli, ['encrypt', self.public_keyfile.name, "--", str(num)])
            assert result.exit_code == 0

    def test_decrypt_positive_integers(self):
        numbers = [0, 1, 2, 5, 10, '1', '10550']

        for num in numbers:
            with tempfile.NamedTemporaryFile() as encfile:
                fname = encfile.name

                self.runner.invoke(cli, [
                    'encrypt', self.public_keyfile.name, str(num), '--output', fname
                ])

                result = self.runner.invoke(cli, [
                    'decrypt', self.private_keyfile.name, fname
                ])
                assert result.exit_code == 0

                assert "{}".format(num) in result.output

    def test_decrypt_signed_integers(self):
        numbers = [0, 1, -1, 10, '1', '-10550']

        for num in numbers:
            with tempfile.NamedTemporaryFile() as encfile:
                fname = encfile.name
                self.runner.invoke(cli, [
                    'encrypt', self.public_keyfile.name, '--output', fname, '--', str(num),
                ])

                result = self.runner.invoke(cli, [
                    'decrypt', self.private_keyfile.name, fname
                ])
                assert result.exit_code == 0

                print(result.output)
                assert "{}".format(num) in result.output

    def test_decrypt_float(self):
        numbers = [0.0, 1.1, -0.0001, 100000.01, '1e-20', '-10550e20']

        for num in numbers:
            with tempfile.NamedTemporaryFile() as encfile:
                fname = encfile.name
                self.runner.invoke(cli, [
                    'encrypt', self.public_keyfile.name, '--output', fname, '--', str(num),
                ])

                with tempfile.NamedTemporaryFile() as outfile:
                    result = self.runner.invoke(cli, [
                        'decrypt', self.private_keyfile.name, fname, '--output', outfile.name
                    ])
                    assert result.exit_code == 0

                    out = outfile.read()
                    self.assertAlmostEqual(float(num), float(out))


class TestConsoleHelpers(TestCase):

    @classmethod
    def setUpClass(cls):
        """Generate a keypair, extract the public key, and encrypt
        a list of numbers

        """
        cls.private_keyfile = tempfile.NamedTemporaryFile()
        cls.public_keyfile = tempfile.NamedTemporaryFile()


        cls.runner = CliRunner()
        cls.runner.invoke(cli, ['genpkey', '--keysize', '256', cls.private_keyfile.name])
        cls.runner.invoke(cli, ['extract', cls.private_keyfile.name, cls.public_keyfile.name])

    def setUp(self):
        self.enc_a_file = tempfile.NamedTemporaryFile()
        self.enc_b_file = tempfile.NamedTemporaryFile()
        self.enc_result_file = tempfile.NamedTemporaryFile()

    def encrypt_and_add(self, a, b):
        self.runner.invoke(cli,
                           ['encrypt', self.public_keyfile.name, '--output', self.enc_a_file.name, '--', str(a)])
        self.runner.invoke(cli,
                           ['encrypt', self.public_keyfile.name, '--output', self.enc_b_file.name, '--', str(b)])

        result = self.runner.invoke(cli, [
            'addenc',
            self.public_keyfile.name,
            self.enc_a_file.name,
            self.enc_b_file.name,
            '--output',
            self.enc_result_file.name
        ])

        assert result.exit_code == 0

        with tempfile.NamedTemporaryFile() as outfile:
            result = self.runner.invoke(cli, [
                'decrypt', self.private_keyfile.name, self.enc_result_file.name, '--output', outfile.name
            ])
            assert result.exit_code == 0

            out = outfile.read()
            return float(out)

    def _a_b_encrypt_helper(self, a, b, operation):
        self.runner.invoke(cli,
                           [
                               'encrypt',
                               self.public_keyfile.name,
                                '--output',
                               self.enc_a_file.name,
                               '--',
                                str(a)
                           ])

        result = self.runner.invoke(cli, [
            operation,
            '--output',
            self.enc_result_file.name,
            self.public_keyfile.name,
            self.enc_a_file.name,
            '--',
            str(b)
        ])

        assert result.exit_code == 0, "Problem carrying out the {} operation".format(operation)

        with tempfile.NamedTemporaryFile() as outfile:
            result = self.runner.invoke(cli, [
                'decrypt', self.private_keyfile.name, self.enc_result_file.name, '--output', outfile.name
            ])
            assert result.exit_code == 0

            out = outfile.read()
            return float(out)

    def encrypt_a_and_add_b(self, a, b):
        return self._a_b_encrypt_helper(a, b, 'add')

    def encrypt_a_and_multiply_b(self, a, b):
        return self._a_b_encrypt_helper(a, b, 'multiply')


class TestConsoleAddition(TestConsoleHelpers):

    def test_addenc_int(self):
        a, b = 12345, 6789
        out = self.encrypt_and_add(a, b)
        self.assertAlmostEqual(float(a + b), float(out))

    def test_add_int(self):
        a, b = 12345, 6789
        out = self.encrypt_a_and_add_b(a, b)
        self.assertAlmostEqual(float(a + b), float(out))

    def test_addenc_large_ints(self):
        """Test adding large integers.
        """
        a, b = int(1.2e10), int(1e15)
        out = self.encrypt_and_add(a, b)
        self.assertAlmostEqual(float(a + b), float(out))

    def test_add_large_ints(self):
        """Test adding large integers.
        """
        a, b = int(1.2e10), int(1e15)
        out = self.encrypt_a_and_add_b(a, b)
        self.assertAlmostEqual(float(a + b), float(out))


    def test_addenc_signed_int(self):
        a, b = 12345, -6789
        out = self.encrypt_and_add(a, b)
        self.assertAlmostEqual(float(a + b), float(out))

    def test_add_signed_int(self):
        a, b = 12345, -6789
        out = self.encrypt_a_and_add_b(a, b)
        self.assertAlmostEqual(float(a + b), float(out))

    def test_addenc_floats(self):
        a, b = 123.45, 67.89
        out = self.encrypt_and_add(a, b)
        self.assertAlmostEqual(float(a + b), float(out))

    def test_add_floats(self):
        a, b = 123.45, 67.89
        out = self.encrypt_a_and_add_b(a, b)
        self.assertAlmostEqual(float(a + b), float(out))

    def test_addenc_large_floats(self):
        """Test adding large integers.
        """
        a, b = 2.3e32, 1.4e32
        out = self.encrypt_and_add(a, b)
        self.assertAlmostEqual(float(a + b), float(out))

    def test_add_large_floats(self):
        """Test adding large integers.
        """
        a, b = 2.3e32, 1.4e32
        out = self.encrypt_a_and_add_b(a, b)
        self.assertAlmostEqual(float(a + b), float(out))


class TestConsoleMultiplication(TestConsoleHelpers):
    """
    Expected to fail until we decide if encrypted numbers with different
    exponents are allowed for this CLI...
    """
    def test_multiply_ints(self):
        a, b = 15, 6
        out = self.encrypt_a_and_multiply_b(a, b)
        self.assertAlmostEqual(int(a * b), int(out))

    def test_multiply_floats(self):
        a, b = 1.2345, 0.6
        out = self.encrypt_a_and_multiply_b(a, b)
        self.assertAlmostEqual(float(a * b), float(out))

    def test_multiply_random_ints(self):
        """
        """
        MAX = 100000000000
        MIN = -MAX

        for _ in range(50):
            a, b = random.randrange(MIN, MAX), random.randrange(MIN, MAX)
            out = self.encrypt_a_and_multiply_b(a, b)
            self.assertAlmostEqual(float(a * b), float(out))


class TestFuzz(TestConsoleHelpers):

    def test_addenc_random_ints(self):
        """Test adding random ints
        """
        MAX = 1000000000000000
        MIN = -MAX

        for _ in range(20):
            a, b = random.randrange(MIN, MAX), random.randrange(MIN, MAX)
            out = self.encrypt_and_add(a, b)
            self.assertAlmostEqual(float(a + b), float(out))

    def test_add_random_ints(self):
        """Test adding random ints
        """
        MAX = 1000000000000000
        MIN = -MAX

        for _ in range(20):
            a, b = random.randrange(MIN, MAX), random.randrange(MIN, MAX)
            out = self.encrypt_a_and_add_b(a, b)
            self.assertAlmostEqual(float(a + b), float(out))

    def test_addenc_random_floats(self):
        """Test adding random floating point numbers from the range [0.0, 1.0)
        """
        for _ in range(20):
            a, b = random.random(), random.random()
            out = self.encrypt_and_add(a, b)
            self.assertAlmostEqual(float(a + b), float(out))

    def test_add_random_floats(self):
        """Test adding random floating point numbers from the range [0.0, 1.0)
        """
        for _ in range(20):
            a, b = random.random(), random.random()
            out = self.encrypt_a_and_add_b(a, b)
            self.assertAlmostEqual(float(a + b), float(out))


    def test_multiply_random_ints(self):
        """
        """
        MAX = 10000
        MIN = -MAX

        for _ in range(20):
            a, b = random.randrange(MIN, MAX), random.randrange(MIN, MAX)
            out = self.encrypt_a_and_multiply_b(a, b)
            self.assertAlmostEqual(float(a * b), float(out))





#!/usr/bin/env python

"""Unittest for maths involving the paillier module."""

# This file is part of pyphe.
#
# Pyphe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyphe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyphe.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import numpy as np

from phe import paillier


class PaillierTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Could move this into setUpModule() if we get too many classes
        cls.public_key, cls.private_key = paillier.generate_paillier_keypair()

        enc_flt = cls.public_key.encrypt

        cls.vec4_1_non_neg = [0.3, 1.7, 6857.6, 1e-6]
        cls.vec4_2         = [-68, 1.8, 34, 1.5e6]

        cls.e_vec4_1 = [enc_flt(x) for x in cls.vec4_1_non_neg]
        cls.e_vec4_2 = [enc_flt(x) for x in cls.vec4_2]


class ArithmeticTest(PaillierTest):

    def testMean(self):
        # Check that we can take an average as good as numpy
        e_mean4_1 = np.mean(self.e_vec4_1)
        self.assertAlmostEqual(np.mean(self.vec4_1_non_neg),
                               self.private_key.decrypt(e_mean4_1))

        emean4_2 = np.mean(self.e_vec4_2)
        self.assertAlmostEqual(np.mean(self.vec4_2),
                               self.private_key.decrypt(emean4_2))

    def testDot(self):
        # Check that our dot product is as good as numpy's
        e_dot_4_2_4_1 = np.dot(self.e_vec4_2, self.vec4_1_non_neg)
        self.assertAlmostEqual(np.dot(self.vec4_2, self.vec4_1_non_neg),
                               self.private_key.decrypt(e_dot_4_2_4_1))


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import, division, print_function

__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]

__title__ = "phe_cuda"
__summary__ = "Acceleration for Partially Homomorphic Encryption library by Cuda "
__uri__ = "https://github.com/gxx777/phe_cuda"

# We use semantic versioning - semver.org
__version__ = "1.0.0"

__author__ = "Xiangxin Guo,WenTian Zhu,Yonglin Wu"
__email__ = "xiangxinguo@mail.ustc.edu.cn"
__license__ = "GPLv3"
__copyright__ = "Copyright 2022 {0}".format(__author__)



from phe.__about__ import *
from phe.encoding import EncodedNumber
from phe.paillier import generate_paillier_keypair
from phe.paillier import EncryptedNumber
from phe.paillier import PaillierPrivateKey, PaillierPublicKey
from phe.paillier import PaillierPrivateKeyring

import phe.util

try:
    import phe.command_line
except ImportError:
    pass



import fractions
import math
import sys


class EncodedNumber(object):
    """Represents a float or int encoded for Paillier encryption.

    For end users, this class is mainly useful for specifying precision
    when adding/multiplying an :class:`EncryptedNumber` by a scalar.

    If you want to manually encode a number for Paillier encryption,
    then use :meth:`encode`, if de-serializing then use
    :meth:`__init__`.


    .. note::
        If working with other Paillier libraries you will have to agree on
        a specific :attr:`BASE` and :attr:`LOG2_BASE` - inheriting from this
        class and overriding those two attributes will enable this.

    Notes:
      Paillier encryption is only defined for non-negative integers less
      than :attr:`PaillierPublicKey.n`. Since we frequently want to use
      signed integers and/or floating point numbers (luxury!), values
      should be encoded as a valid integer before encryption.

      The operations of addition and multiplication [1]_ must be
      preserved under this encoding. Namely:

      1. Decode(Encode(a) + Encode(b)) = a + b
      2. Decode(Encode(a) * Encode(b)) = a * b

      for any real numbers a and b.

      Representing signed integers is relatively easy: we exploit the
      modular arithmetic properties of the Paillier scheme. We choose to
      represent only integers between
      +/-:attr:`~PaillierPublicKey.max_int`, where `max_int` is
      approximately :attr:`~PaillierPublicKey.n`/3 (larger integers may
      be treated as floats). The range of values between `max_int` and
      `n` - `max_int` is reserved for detecting overflows. This encoding
      scheme supports properties #1 and #2 above.

      Representing floating point numbers as integers is a harder task.
      Here we use a variant of fixed-precision arithmetic. In fixed
      precision, you encode by multiplying every float by a large number
      (e.g. 1e6) and rounding the resulting product. You decode by
      dividing by that number. However, this encoding scheme does not
      satisfy property #2 above: upon every multiplication, you must
      divide by the large number. In a Paillier scheme, this is not
      possible to do without decrypting. For some tasks, this is
      acceptable or can be worked around, but for other tasks this can't
      be worked around.

      In our scheme, the "large number" is allowed to vary, and we keep
      track of it. It is:

        :attr:`BASE` ** :attr:`exponent`

      One number has many possible encodings; this property can be used
      to mitigate the leak of information due to the fact that
      :attr:`exponent` is never encrypted.

      For more details, see :meth:`encode`.

    .. rubric:: Footnotes

    ..  [1] Technically, since Paillier encryption only supports
      multiplication by a scalar, it may be possible to define a
      secondary encoding scheme `Encode'` such that property #2 is
      relaxed to:

        Decode(Encode(a) * Encode'(b)) = a * b

      We don't do this.


    Args:
      public_key (PaillierPublicKey): public key for which to encode
        (this is necessary because :attr:`~PaillierPublicKey.max_int`
        varies)
      encoding (int): The encoded number to store. Must be positive and
        less than :attr:`~PaillierPublicKey.max_int`.
      exponent (int): Together with :attr:`BASE`, determines the level
        of fixed-precision used in encoding the number.

    Attributes:
      public_key (PaillierPublicKey): public key for which to encode
        (this is necessary because :attr:`~PaillierPublicKey.max_int`
        varies)
      encoding (int): The encoded number to store. Must be positive and
        less than :attr:`~PaillierPublicKey.max_int`.
      exponent (int): Together with :attr:`BASE`, determines the level
        of fixed-precision used in encoding the number.
    """
    BASE = 16
    """Base to use when exponentiating. Larger `BASE` means
    that :attr:`exponent` leaks less information. If you vary this,
    you'll have to manually inform anyone decoding your numbers.
    """
    LOG2_BASE = math.log(BASE, 2)
    FLOAT_MANTISSA_BITS = sys.float_info.mant_dig

    def __init__(self, public_key, encoding, exponent):
        self.public_key = public_key
        self.encoding = encoding
        self.exponent = exponent

    @classmethod
    def encode(cls, public_key, scalar, precision=None, max_exponent=None):
        """Return an encoding of an int or float.

        This encoding is carefully chosen so that it supports the same
        operations as the Paillier cryptosystem.

        If *scalar* is a float, first approximate it as an int, `int_rep`:

            scalar = int_rep * (:attr:`BASE` ** :attr:`exponent`),

        for some (typically negative) integer exponent, which can be
        tuned using *precision* and *max_exponent*. Specifically,
        :attr:`exponent` is chosen to be equal to or less than
        *max_exponent*, and such that the number *precision* is not
        rounded to zero.

        Having found an integer representation for the float (or having
        been given an int `scalar`), we then represent this integer as
        a non-negative integer < :attr:`~PaillierPublicKey.n`.

        Paillier homomorphic arithemetic works modulo
        :attr:`~PaillierPublicKey.n`. We take the convention that a
        number x < n/3 is positive, and that a number x > 2n/3 is
        negative. The range n/3 < x < 2n/3 allows for overflow
        detection.

        Args:
          public_key (PaillierPublicKey): public key for which to encode
            (this is necessary because :attr:`~PaillierPublicKey.n`
            varies).
          scalar: an int or float to be encrypted.
            If int, it must satisfy abs(*value*) <
            :attr:`~PaillierPublicKey.n`/3.
            If float, it must satisfy abs(*value* / *precision*) <<
            :attr:`~PaillierPublicKey.n`/3
            (i.e. if a float is near the limit then detectable
            overflow may still occur)
          precision (float): Choose exponent (i.e. fix the precision) so
            that this number is distinguishable from zero. If `scalar`
            is a float, then this is set so that minimal precision is
            lost. Lower precision leads to smaller encodings, which
            might yield faster computation.
          max_exponent (int): Ensure that the exponent of the returned
            `EncryptedNumber` is at most this.

        Returns:
          EncodedNumber: Encoded form of *scalar*, ready for encryption
          against *public_key*.
        """
        # Calculate the maximum exponent for desired precision
        if precision is None:
            if isinstance(scalar, int):
                prec_exponent = 0
            elif isinstance(scalar, float):
                # Encode with *at least* as much precision as the python float
                # What's the base-2 exponent on the float?
                bin_flt_exponent = math.frexp(scalar)[1]

                # What's the base-2 exponent of the least significant bit?
                # The least significant bit has value 2 ** bin_lsb_exponent
                bin_lsb_exponent = bin_flt_exponent - cls.FLOAT_MANTISSA_BITS

                # What's the corresponding base BASE exponent? Round that down.
                prec_exponent = math.floor(bin_lsb_exponent / cls.LOG2_BASE)
            else:
                raise TypeError("Don't know the precision of type %s."
                                % type(scalar))
        else:
            prec_exponent = math.floor(math.log(precision, cls.BASE))

        # Remember exponents are negative for numbers < 1.
        # If we're going to store numbers with a more negative
        # exponent than demanded by the precision, then we may
        # as well bump up the actual precision.
        if max_exponent is None:
            exponent = prec_exponent
        else:
            exponent = min(max_exponent, prec_exponent)

        # Use rationals instead of floats to avoid overflow.
        int_rep = round(fractions.Fraction(scalar)
                        * fractions.Fraction(cls.BASE) ** -exponent)

        if abs(int_rep) > public_key.max_int:
            raise ValueError('Integer needs to be within +/- %d but got %d'
                             % (public_key.max_int, int_rep))

        # Wrap negative numbers by adding n
        return cls(public_key, int_rep % public_key.n, exponent)

    def decode(self):
        """Decode plaintext and return the result.

        Returns:
          an int or float: the decoded number. N.B. if the number
            returned is an integer, it will not be of type float.

        Raises:
          OverflowError: if overflow is detected in the decrypted number.
        """
        if self.encoding >= self.public_key.n:
            # Should be mod n
            raise ValueError('Attempted to decode corrupted number')
        elif self.encoding <= self.public_key.max_int:
            # Positive
            mantissa = self.encoding
        elif self.encoding >= self.public_key.n - self.public_key.max_int:
            # Negative
            mantissa = self.encoding - self.public_key.n
        else:
            raise OverflowError('Overflow detected in decrypted number')

        if self.exponent >= 0:
            # Integer multiplication. This is exact.
            return mantissa * self.BASE ** self.exponent
        else:
            # BASE ** -e is an integer, so below is a division of ints.
            # Not coercing mantissa to float prevents some overflows.
            try:
                return mantissa / self.BASE ** -self.exponent
            except OverflowError as e:
                raise OverflowError(
                    'decoded result too large for a float') from e

    def decrease_exponent_to(self, new_exp):
        """Return an `EncodedNumber` with same value but lower exponent.

        If we multiply the encoded value by :attr:`BASE` and decrement
        :attr:`exponent`, then the decoded value does not change. Thus
        we can almost arbitrarily ratchet down the exponent of an
        :class:`EncodedNumber` - we only run into trouble when the encoded
        integer overflows. There may not be a warning if this happens.

        This is necessary when adding :class:`EncodedNumber` instances,
        and can also be useful to hide information about the precision
        of numbers - e.g. a protocol can fix the exponent of all
        transmitted :class:`EncodedNumber` to some lower bound(s).

        Args:
          new_exp (int): the desired exponent.

        Returns:
          EncodedNumber: Instance with the same value and desired
            exponent.

        Raises:
          ValueError: You tried to increase the exponent, which can't be
            done without decryption.
        """
        if new_exp > self.exponent:
            raise ValueError('New exponent %i should be more negative than'
                             'old exponent %i' % (new_exp, self.exponent))
        factor = pow(self.BASE, self.exponent - new_exp)
        new_enc = self.encoding * factor % self.public_key.n
        return self.__class__(self.public_key, new_enc, new_exp)



#!/usr/bin/env python3.4

# This file is part of pyphe.
#
# Pyphe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyphe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyphe.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import random
import math
try:
    from math import gcd  # new in Python 3.5
except ImportError:
    from fractions import gcd  # deprecated since Python 3.5

from phe import util


class PaillierUtilTest(unittest.TestCase):

    def testPowMod(self):
        self.assertEqual(util.powmod(5, 3, 3), 2)
        self.assertEqual(util.powmod(2, 10, 1000), 24)

    def testInvert(self):
        p = 101
        for i in range(1, p):
            iinv = util.invert(i, p)
            self.assertEqual((iinv * i) % p, 1)

    def testInvertNonPrime(self):
        a = 3
        p = 4
        self.assertEqual(a * util.invert(a, p) % p, 1)

    def testPrimeOverN(self):
        self.assertIn(util.getprimeover(3), {5, 7, 11, 13})
        # Test we get at least a N-bit prime
        for n in range(2, 50):
            p = util.getprimeover(n)
            self.assertGreaterEqual(p, 1 << (n-1))

    def testIsqrt(self):
        for _ in range(100):
            n = random.randint(2, 10000000)
            nsq = n*n
            self.assertEqual(int(math.floor(math.sqrt(n))), util.isqrt(n))
            self.assertEqual(util.isqrt(nsq), util.improved_i_sqrt(nsq))


# same tests as above, but with gmpy2 and Crypto libraries disabled
class PaillierUtilFallbacksTest(PaillierUtilTest):

    def setUp(self):
        # save presence of libraries
        self.HAVE_GMP = util.HAVE_GMP
        self.HAVE_CRYPTO = util.HAVE_CRYPTO
        # disable libraties
        util.HAVE_GMP = False
        util.HAVE_CRYPTO = False

    def tearDown(self):
        # restore presence of libraries
        util.HAVE_GMP = self.HAVE_GMP
        util.HAVE_CRYPTO = self.HAVE_CRYPTO

    def testExtendedEuclieanAlgorithm(self):
        # from <https://en.wikipedia.org/wiki/Extended_Euclidean_algorithm>
        self.assertEqual(util.extended_euclidean_algorithm(240, 46), (2, -9, 47))

        # tests with arbirary values
        for a, b in [(77, 99), (45, 127)]:  # non-coprime pair, coprime pair
            r, s, t = util.extended_euclidean_algorithm(a, b)
            self.assertEqual(r, s*a + t*b)
            self.assertEqual(r, gcd(a, b))

    def testMillerRabin(self):
        a = 2  # witness, enough by itself for checking n < 2047
        self.assertFalse(util.miller_rabin(4, a))
        self.assertTrue(util.miller_rabin(127, a))
        composite = util.first_primes[-1] * util.first_primes[-2]
        self.assertFalse(util.miller_rabin(composite, a))

    def testIsPrime(self):
        self.assertTrue(util.is_prime(17881))  # first not in first_primes
        self.assertFalse(util.is_prime(-17881))

        self.assertFalse(util.is_prime(-4))
        self.assertFalse(util.is_prime(-2))
        self.assertFalse(util.is_prime(-1))
        self.assertFalse(util.is_prime(0))
        self.assertFalse(util.is_prime(1))
        self.assertTrue(util.is_prime(2))
        self.assertTrue(util.is_prime(3))

        # same tests as for miller_rabin()
        self.assertFalse(util.is_prime(4))
        self.assertTrue(util.is_prime(127))
        composite = util.first_primes[-1] * util.first_primes[-2]
        self.assertFalse(util.is_prime(composite))


class Base64UtilTest(unittest.TestCase):

    def testEncodeDecodePositiveNonZeroInt(self):
        for a in range(1, 1000000, 100):

            self.assertEqual(a, util.base64_to_int(util.int_to_base64(a)))

    def testFailToEncodeZero(self):
        with self.assertRaises(AssertionError):
            util.int_to_base64(0)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
#-*- coding:utf-8 -*-

#############################################
# File Name: setup.py
# Author: mage
# Created Time:  2022-12-15 22:17:34
#############################################


from setuptools import setup, find_packages

setup(
    name = "phe_cuda",
    version = "0.1.0",
    keywords = ("pip", "cuda","phe", "paillier", "mage"),
    description = "Acceleration Paillier by cuda",
    long_description = "Acceleration for Partially Homomorphic Encryption library by Cuda ",
    license = "MIT Licence",

    url = "https://github.com/gxx777/phe_cuda",
    author = "Xiangxin Guo",
    author_email = "xiangxinguo@mail.ustc.edu.cn",

    packages = find_packages(),
    include_package_data = True,
    platforms = "any",
    install_requires = []
)



#!/usr/bin/env python3
from __future__ import print_function
import datetime
import json

import click
import phe

__author__ = 'brian'


def log(m, color='red'):
    click.echo(click.style(m, fg=color), err=True)


@click.group("pheutil")
@click.version_option('1.0-alpha')
@click.option('--verbose', '-v', is_flag=True,
              help='Enables verbose mode.')
def cli(verbose=False):
    """CLI for interacting with python-paillier
    """

@cli.command("genpkey")
@click.argument('output', type=click.File('w'))
@click.option("--keysize", type=int, default=2048,
              help="The keysize in bits. Defaults to 2048")
@click.option("--id", type=str, default=None,
              help="Add an identifying comment to the key")
def generate_keypair(keysize, id, output):
    """Generate a paillier private key.

    Output as JWK to given output file. Use "-" to output the private key to
    stdout. See the extract command to extract the public component of the
    private key.

    Note:
        The default ID text includes the current time.
    """
    log("Generating a paillier keypair with keysize of {}".format(keysize))
    pub, priv = phe.generate_paillier_keypair(n_length=keysize)

    log("Keys generated")

    date = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    jwk_public = {
        'kty': "DAJ",
        'alg': "PAI-GN1",
        "key_ops": ["encrypt"],
        'n': phe.util.int_to_base64(pub.n),
        'kid': "Paillier public key generated by pheutil on {}".format(date)
    }

    jwk_private = {
        'kty': "DAJ",
        'key_ops': ["decrypt"],
        'p': phe.util.int_to_base64(priv.p),
        'q': phe.util.int_to_base64(priv.q),
        'pub': jwk_public,
        'kid': "Paillier private key generated by pheutil on {}".format(date)
    }

    json.dump(jwk_private, output)
    output.write('\n')

    log("Private key written to {}".format(output.name))


@cli.command()
@click.argument('input', type=click.File('r'))
@click.argument('output', type=click.File('w'))
def extract(input, output):
    """Extract public key from private key.

    Given INPUT a private paillier key file as generated by generate, extract the
    public key portion to OUTPUT.

    Use "-" to output to stdout.
    """
    log("Loading paillier keypair")
    priv = json.load(input)
    error_msg = "Invalid private key"
    assert 'pub' in priv, error_msg
    assert priv['kty'] == 'DAJ', error_msg
    json.dump(priv['pub'], output)
    output.write('\n')
    log("Public key written to {}".format(output.name))



@cli.command()
@click.argument('public', type=click.File('r'))
@click.argument('plaintext', type=str)
@click.option('--output', type=click.File('w'),
              help="Save to file instead of stdout")
def encrypt(public, plaintext, output=None):
    """Encrypt a number with public key.

    The PLAINTEXT input will be interpreted as a floating point number.

    Output will be a JSON object with a "v" attribute containing the
    ciphertext as a string, and "e" the exponent as a Number,
    where possible fixed at -32.

    Note if you are passing a negative number to encrypt, you will
    need to include a "--" between the public key and your plaintext.
    """
    num = float(plaintext)

    log("Loading public key")
    publickeydata = json.load(public)
    pub = load_public_key(publickeydata)

    log("Encrypting: {:+.16f}".format(num))
    enc = pub.encrypt(num)
    serialised = serialise_encrypted(enc)
    print(serialised, file=output)


def serialise_encrypted(enc):
    if enc.exponent > -32:
        enc = enc.decrease_exponent_to(-32)
        assert enc.exponent == -32
    else:
        log("Exponent is less than -32")

    obj = json.dumps({
        "v": str(enc.ciphertext()),
        "e": enc.exponent
    })
    return obj


@cli.command()
@click.argument('private', type=click.File('r'))
@click.argument('ciphertext', type=click.File('r'))
@click.option('--output', type=click.File('w'),
              help="Save to file instead of stdout")
def decrypt(private, ciphertext, output):
    """Decrypt ciphertext with private key.

    Requires PRIVATE key file and the CIPHERTEXT encrypted with
    the corresponding public key.
    """
    privatekeydata = json.load(private)
    assert 'pub' in privatekeydata
    pub = load_public_key(privatekeydata['pub'])

    log("Loading private key")
    private_key_error = "Invalid private key"
    assert 'key_ops' in privatekeydata, private_key_error
    assert "decrypt" in privatekeydata['key_ops'], private_key_error
    assert 'p' in privatekeydata, private_key_error
    assert 'q' in privatekeydata, private_key_error
    assert privatekeydata['kty'] == 'DAJ', private_key_error

    _p = phe.util.base64_to_int(privatekeydata['p'])
    _q = phe.util.base64_to_int(privatekeydata['q'])

    private_key = phe.PaillierPrivateKey(pub, _p, _q)

    log("Decrypting ciphertext")
    enc = load_encrypted_number(ciphertext, pub)
    out = private_key.decrypt(enc)
    print(out, file=output)


@cli.command("addenc")
@click.argument('public', type=click.File('r'))
@click.argument('encrypted_a', type=click.File('r'))
@click.argument('encrypted_b', type=click.File('r'))
@click.option('--output', type=click.File('w'),
              help="Save to file instead of stdout")
def add_encrypted(public, encrypted_a, encrypted_b, output):
    """Add two encrypted numbers together.

    """
    log("Loading public key")
    publickeydata = json.load(public)
    pub = load_public_key(publickeydata)

    log("Loading first encrypted number")
    enc_a = load_encrypted_number(encrypted_a, pub)

    log("Loading second encrypted number")
    enc_b = load_encrypted_number(encrypted_b, pub)

    log("Adding encrypted numbers together")

    enc_result = enc_a + enc_b
    serialised_result = serialise_encrypted(enc_result)
    print(serialised_result, file=output)


@cli.command("add")
@click.argument('public', type=click.File('r'))
@click.argument('encrypted', type=click.File('r'))
@click.argument('plaintext', type=str)
@click.option('--output', type=click.File('w'),
              help="Save to file instead of stdout")
def add_encrypted_to_plaintext(public, encrypted, plaintext, output):
    """Add encrypted number to unencrypted number.

    Requires a PUBLIC key file, a number ENCRYPTED with that public key
    also as a file, and the PLAINTEXT number to add.

    Creates a new encrypted number.
    """
    log("Loading public key")
    publickeydata = json.load(public)
    pub = load_public_key(publickeydata)

    log("Loading encrypted number")
    enc = load_encrypted_number(encrypted, pub)

    log("Loading unencrypted number")
    num = float(plaintext)

    log("Adding")
    enc_result = enc + num
    serialised_result = serialise_encrypted(enc_result)
    print(serialised_result, file=output)


@cli.command("multiply")
@click.argument('public', type=click.File('r'))
@click.argument('encrypted', type=click.File('r'))
@click.argument('plaintext', type=str)
@click.option('--output', type=click.File('w'),
              help="Save to file instead of stdout")
def multiply_encrypted_to_plaintext(public, encrypted, plaintext, output):
    """Multiply encrypted num with unencrypted num.

    Requires a PUBLIC key file, a number ENCRYPTED with that public key
    also as a file, and the PLAINTEXT number to multiply.

    Creates a new encrypted number.
    """
    log("Loading public key")
    publickeydata = json.load(public)
    pub = load_public_key(publickeydata)

    log("Loading encrypted number")
    enc = load_encrypted_number(encrypted, pub)

    log("Loading unencrypted number")
    num = float(plaintext)

    log("Multiplying")
    enc_result = enc * num

    serialised_result = serialise_encrypted(enc_result)
    print(serialised_result, file=output)


def load_public_key(public_key_data):
    error_msg = "Invalid public key"
    assert 'alg' in public_key_data, error_msg
    assert public_key_data['alg'] == 'PAI-GN1', error_msg
    assert public_key_data['kty'] == 'DAJ', error_msg

    n = phe.util.base64_to_int(public_key_data['n'])
    pub = phe.PaillierPublicKey(n)
    return pub


def load_encrypted_number(enc_number_file, pub):
    ciphertext_data = json.load(enc_number_file)
    assert 'v' in ciphertext_data
    assert 'e' in ciphertext_data

    enc = phe.EncryptedNumber(pub,
                              int(ciphertext_data['v']),
                              exponent=ciphertext_data['e']
                              )
    return enc


if __name__ == "__main__":
    cli()
